# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb.web.views.basecontrollers unit tests"""

import time
from urllib.parse import urlsplit, urlunsplit, urljoin, parse_qs

import lxml
from cubicweb import Binary, NoSelectableObject, ValidationError, transaction as tx
from cubicweb.devtools import BASE_URL
from cubicweb.predicates import is_instance
from cubicweb.schema import RRQLExpression
from cubicweb.server.hook import Hook, Operation
from cubicweb.server.session import Connection
from cubicweb.uilib import rql_for_eid
from cubicweb.utils import json_dumps
from logilab.common.testlib import unittest_main

from cubicweb_web import Redirect, RemoteCallFailed, http_headers, formfields as ff
from cubicweb_web.devtools.testlib import WebCWTC, PyramidWebCWTC
from cubicweb_web.views.ajaxcontroller import ajaxfunc, AjaxFunction
from cubicweb_web.views.autoform import get_pending_inserts, get_pending_deletes


class ViewControllerTC(PyramidWebCWTC):
    def test_view_ctrl_with_valid_cache_headers(self):
        now = time.time()
        resp = self.webapp.get(urljoin(BASE_URL, "manage"))
        self.assertEqual(resp.etag, "manage/guests")
        self.assertEqual(resp.status_code, 200)
        self.assertGreaterEqual(
            http_headers.parseDateTime(resp.headers["Last-Modified"]), int(now)
        )
        cache_headers = {
            "if-modified-since": resp.headers["Last-Modified"],
            "if-none-match": resp.etag,
        }
        resp = self.webapp.get(urljoin(BASE_URL, "manage"), headers=cache_headers)
        self.assertEqual(resp.status_code, 304)
        self.assertEqual(len(resp.body), 0)


def req_form(user):
    return {
        "eid": [str(user.eid)],
        "_cw_entity_fields:%s" % user.eid: "_cw_generic_field",
        "__type:%s" % user.eid: user.__regid__,
    }


class EditControllerTC(PyramidWebCWTC):
    def setUp(self):
        WebCWTC.setUp(self)
        self.assertIn(
            "users", self.schema.entity_schema_for("CWGroup").get_groups("read")
        )

    def tearDown(self):
        WebCWTC.tearDown(self)
        self.assertIn(
            "users", self.schema.entity_schema_for("CWGroup").get_groups("read")
        )

    def test_noparam_edit(self):
        """check behaviour of this controller without any form parameter"""
        with self.admin_access.web_request() as req:
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            self.assertEqual(cm.exception.errors, {None: "no selected entities"})

    def test_validation_unique(self):
        """test creation of two linked entities"""
        with self.admin_access.web_request() as req:
            req.form = {
                "eid": "X",
                "__type:X": "CWUser",
                "_cw_entity_fields:X": "login-subject,upassword-subject",
                "login-subject:X": "admin",
                "upassword-subject:X": "toto",
                "upassword-subject-confirm:X": "toto",
            }
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            cm.exception.translate(str)
            expected = {
                "": "some relations violate a unicity constraint",
                "login": "login is part of violated unicity constraint",
            }
            self.assertEqual(cm.exception.errors, expected)

    def test_simultaneous_edition_only_one_commit(self):
        """Allow two simultaneous edit view of the same entity as long as only one commits"""
        with self.admin_access.web_request() as req:
            e = req.create_entity("BlogEntry", title="cubicweb.org", content="hop")
            expected_path = e.rest_path()
            req.cnx.commit()
            form = self.vreg["views"].select("edition", req, rset=e.as_rset(), row=0)
            html_form = lxml.html.fromstring(form.render(w=None, action="edit")).forms[
                0
            ]

        with self.admin_access.web_request():
            self.vreg["views"].select("edition", req, rset=e.as_rset(), row=0)

        with self.admin_access.web_request(**dict(html_form.form_values())) as req:
            path, args = self.expect_redirect_handle_request(req, path="edit")
            self.assertEqual(path, expected_path)

    def test_simultaneous_edition_refuse_second_commit(self):
        """Disallow committing changes to an entity edited in between"""
        with self.admin_access.web_request() as req:
            e = req.create_entity("BlogEntry", title="cubicweb.org", content="hop")
            eid = e.eid
            req.cnx.commit()
            form = self.vreg["views"].select("edition", req, rset=e.as_rset(), row=0)
            html_form = lxml.html.fromstring(form.render(w=None, action="edit")).forms[
                0
            ]

        with self.admin_access.web_request() as req2:
            e = req2.entity_from_eid(eid)
            e.cw_set(content="hip")
            req2.cnx.commit()

        form_field_name = "content-subject:%d" % eid
        form_values = dict(html_form.form_values())
        assert form_field_name in form_values
        form_values[form_field_name] = "yep"
        with self.admin_access.web_request(**form_values) as req:
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            reported_eid, dict_info = cm.exception.args
            self.assertEqual(reported_eid, eid)
            self.assertIn(None, dict_info)
            self.assertIn("has changed since you started to edit it.", dict_info[None])

    def test_user_editing_itself(self):
        """checking that a manager user can edit itself"""
        with self.admin_access.web_request() as req:
            user = req.user
            groupeids = [
                eid
                for eid, in req.execute(
                    "CWGroup G WHERE G name " 'in ("managers", "users")'
                )
            ]
            groups = [str(eid) for eid in groupeids]
            eid = str(user.eid)
            req.form = {
                "eid": eid,
                "__type:" + eid: "CWUser",
                "_cw_entity_fields:"
                + eid: "login-subject,firstname-subject,surname-subject,in_group-subject",
                "login-subject:" + eid: str(user.login),
                "surname-subject:" + eid: "Th\xe9nault",
                "firstname-subject:" + eid: "Sylvain",
                "in_group-subject:" + eid: groups,
            }
            self.expect_redirect_handle_request(req, "edit")
            e = req.execute("Any X WHERE X eid %(x)s", {"x": user.eid}).get_entity(0, 0)
            self.assertEqual(e.firstname, "Sylvain")
            self.assertEqual(e.surname, "Th\xe9nault")
            self.assertEqual(e.login, user.login)
            self.assertEqual([g.eid for g in e.in_group], groupeids)

    def test_user_can_change_its_password(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, "user")
            cnx.commit()
        with self.new_access("user").web_request() as req:
            eid = str(req.user.eid)
            req.form = {
                "eid": eid,
                "__maineid": eid,
                "__type:" + eid: "CWUser",
                "_cw_entity_fields:" + eid: "upassword-subject",
                "upassword-subject:" + eid: "tournicoton",
                "upassword-subject-confirm:" + eid: "tournicoton",
            }
            path, params = self.expect_redirect_handle_request(req, "edit")
            req.cnx.commit()  # commit to check we don't get late validation error for instance
            self.assertEqual(path, "cwuser/user")
            self.assertNotIn("vid", params)

    def test_user_editing_itself_no_relation(self):
        """checking we can edit an entity without specifying some required
        relations (meaning no changes)
        """
        with self.admin_access.web_request() as req:
            user = req.user
            groupeids = [g.eid for g in user.in_group]
            eid = str(user.eid)
            req.form = {
                "eid": eid,
                "__type:" + eid: "CWUser",
                "_cw_entity_fields:"
                + eid: "login-subject,firstname-subject,surname-subject",
                "login-subject:" + eid: str(user.login),
                "firstname-subject:" + eid: "Th\xe9nault",
                "surname-subject:" + eid: "Sylvain",
            }
            self.expect_redirect_handle_request(req, "edit")
            e = req.execute("Any X WHERE X eid %(x)s", {"x": user.eid}).get_entity(0, 0)
            self.assertEqual(e.login, user.login)
            self.assertEqual(e.firstname, "Th\xe9nault")
            self.assertEqual(e.surname, "Sylvain")
            self.assertEqual([g.eid for g in e.in_group], groupeids)
            self.assertEqual(e.cw_adapt_to("IWorkflowable").state, "activated")

    def test_create_multiple_linked(self):
        with self.admin_access.web_request() as req:
            gueid = req.execute('CWGroup G WHERE G name "users"')[0][0]
            req.form = {
                "eid": ["X", "Y"],
                "__maineid": "X",
                "__type:X": "CWUser",
                "_cw_entity_fields:X": "login-subject,upassword-subject,surname-subject,in_group-subject",
                "login-subject:X": "adim",
                "upassword-subject:X": "toto",
                "upassword-subject-confirm:X": "toto",
                "surname-subject:X": "Di Mascio",
                "in_group-subject:X": str(gueid),
                "__type:Y": "EmailAddress",
                "_cw_entity_fields:Y": "address-subject,use_email-object",
                "address-subject:Y": "dima@logilab.fr",
                "use_email-object:Y": "X",
            }
            path, _params = self.expect_redirect_handle_request(req, "edit")
            # should be redirected on the created person
            self.assertEqual(path, "cwuser/adim")
            e = req.execute('Any P WHERE P surname "Di Mascio"').get_entity(0, 0)
            self.assertEqual(e.surname, "Di Mascio")
            email = e.use_email[0]
            self.assertEqual(email.address, "dima@logilab.fr")

    def test_create_mandatory_inlined(self):
        with self.admin_access.web_request() as req:
            req.form = {
                "eid": ["X", "Y"],
                "__maineid": "X",
                "__type:X": "Salesterm",
                "_cw_entity_fields:X": "",
                "__type:Y": "File",
                "_cw_entity_fields:Y": "data-subject,described_by_test-object",
                "data-subject:Y": ("coucou.txt", Binary(b"coucou")),
                "described_by_test-object:Y": "X",
            }
            path, _params = self.expect_redirect_handle_request(req, "edit")
            self.assertTrue(path.startswith("salesterm/"), path)
            eid = path.split("/")[1]
            salesterm = req.entity_from_eid(eid)
            # The NOT NULL constraint of mandatory relation implies that the File
            # must be created before the Salesterm, otherwise Salesterm insertion
            # will fail.
            # NOTE: sqlite does have NOT NULL constraint, unlike Postgres so the
            # insertion does not fail and we have to check dumbly that File is
            # created before.
            self.assertGreater(salesterm.eid, salesterm.described_by_test[0].eid)

    def test_create_mandatory_inlined2(self):
        with self.admin_access.web_request() as req:
            req.form = {
                "eid": ["X", "Y"],
                "__maineid": "X",
                "__type:X": "Salesterm",
                "_cw_entity_fields:X": "described_by_test-subject",
                "described_by_test-subject:X": "Y",
                "__type:Y": "File",
                "_cw_entity_fields:Y": "data-subject",
                "data-subject:Y": ("coucou.txt", Binary(b"coucou")),
            }
            path, _params = self.expect_redirect_handle_request(req, "edit")
            self.assertTrue(path.startswith("salesterm/"), path)
            eid = path.split("/")[1]
            salesterm = req.entity_from_eid(eid)
            # The NOT NULL constraint of mandatory relation implies that the File
            # must be created before the Salesterm, otherwise Salesterm insertion
            # will fail.
            # NOTE: sqlite does have NOT NULL constraint, unlike Postgres so the
            # insertion does not fail and we have to check dumbly that File is
            # created before.
            self.assertGreater(salesterm.eid, salesterm.described_by_test[0].eid)

    def test_edit_mandatory_inlined3_object(self):
        # non regression test for #3120495. Without the fix, leads to
        # "unhashable type: 'list'" error
        with self.admin_access.web_request() as req:
            attribute_eid_1 = req.execute("Any X WHERE X is CWAttribute")[0][0]
            attribute_eid_2 = req.execute("Any X WHERE X is CWAttribute")[1][0]
            cwrelation = str(req.execute('CWEType X WHERE X name "CWSource"')[0][0])
            req.form = {
                "eid": [cwrelation],
                "__maineid": cwrelation,
                "__type:" + cwrelation: "CWEType",
                "_cw_entity_fields:" + cwrelation: "to_entity-object",
                "to_entity-object:" + cwrelation: [attribute_eid_1, attribute_eid_2],
            }
            with req.cnx.deny_all_hooks_but():
                path, _params = self.expect_redirect_handle_request(req, "edit")
            self.assertTrue(path.startswith("cwetype/CWSource"), path)

    def test_edit_multiple_linked(self):
        with self.admin_access.web_request() as req:
            peid = str(self.create_user(req, "adim").eid)
            req.form = {
                "eid": [peid, "Y"],
                "__maineid": peid,
                "__type:" + peid: "CWUser",
                "_cw_entity_fields:" + peid: "surname-subject",
                "surname-subject:" + peid: "Di Masci",
                "__type:Y": "EmailAddress",
                "_cw_entity_fields:Y": "address-subject,use_email-object",
                "address-subject:Y": "dima@logilab.fr",
                "use_email-object:Y": peid,
            }
            path, _params = self.expect_redirect_handle_request(req, "edit")
            # should be redirected on the created person
            self.assertEqual(path, "cwuser/adim")
            e = req.execute('Any P WHERE P surname "Di Masci"').get_entity(0, 0)
            email = e.use_email[0]
            self.assertEqual(email.address, "dima@logilab.fr")

            # with self.admin_access.web_request() as req:
            emaileid = str(email.eid)
            req.form = {
                "eid": [peid, emaileid],
                "__type:" + peid: "CWUser",
                "_cw_entity_fields:" + peid: "surname-subject",
                "surname-subject:" + peid: "Di Masci",
                "__type:" + emaileid: "EmailAddress",
                "_cw_entity_fields:" + emaileid: "address-subject,use_email-object",
                "address-subject:" + emaileid: "adim@logilab.fr",
                "use_email-object:" + emaileid: peid,
            }
            self.expect_redirect_handle_request(req, "edit")
            email.cw_clear_all_caches()
            self.assertEqual(email.address, "adim@logilab.fr")

    def test_password_confirm(self):
        """test creation of two linked entities"""
        with self.admin_access.web_request() as req:
            user = req.user
            req.form = {
                "eid": "X",
                "__cloned_eid:X": str(user.eid),
                "__type:X": "CWUser",
                "_cw_entity_fields:X": "login-subject,upassword-subject",
                "login-subject:X": "toto",
                "upassword-subject:X": "toto",
            }
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            self.assertEqual(
                {"upassword-subject": "password and confirmation don't match"},
                cm.exception.errors,
            )
            req.form = {
                "__cloned_eid:X": str(user.eid),
                "eid": "X",
                "__type:X": "CWUser",
                "_cw_entity_fields:X": "login-subject,upassword-subject",
                "login-subject:X": "toto",
                "upassword-subject:X": "toto",
                "upassword-subject-confirm:X": "tutu",
            }
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            self.assertEqual(
                {"upassword-subject": "password and confirmation don't match"},
                cm.exception.errors,
            )

    def test_interval_bound_constraint_success(self):
        with self.admin_access.repo_cnx() as cnx:
            feid = cnx.execute(
                'INSERT File X: X data_name "toto.txt", X data %(data)s',
                {"data": Binary(b"yo")},
            )[0][0]
            cnx.commit()

        with self.admin_access.web_request(rollbackfirst=True) as req:
            req.form = {
                "eid": ["X"],
                "__type:X": "Salesterm",
                "_cw_entity_fields:X": "amount-subject,described_by_test-subject",
                "amount-subject:X": "-10",
                "described_by_test-subject:X": str(feid),
            }
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            cm.exception.translate(str)
            self.assertEqual(
                {"amount-subject": "value -10 must be >= 0"}, cm.exception.errors
            )

        with self.admin_access.web_request(rollbackfirst=True) as req:
            req.form = {
                "eid": ["X"],
                "__type:X": "Salesterm",
                "_cw_entity_fields:X": "amount-subject,described_by_test-subject",
                "amount-subject:X": "110",
                "described_by_test-subject:X": str(feid),
            }
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            cm.exception.translate(str)
            self.assertEqual(
                cm.exception.errors, {"amount-subject": "value 110 must be <= 100"}
            )

        with self.admin_access.web_request(rollbackfirst=True) as req:
            req.form = {
                "eid": ["X"],
                "__type:X": "Salesterm",
                "_cw_entity_fields:X": "amount-subject,described_by_test-subject",
                "amount-subject:X": "10",
                "described_by_test-subject:X": str(feid),
            }
            self.expect_redirect_handle_request(req, "edit")
            # should be redirected on the created
            # eid = params['rql'].split()[-1]
            e = req.execute("Salesterm X").get_entity(0, 0)
            self.assertEqual(e.amount, 10)

    def test_interval_bound_constraint_validateform(self):
        """Test the FormValidatorController controller on entity with
        constrained attributes"""
        with self.admin_access.repo_cnx() as cnx:
            feid = cnx.execute(
                'INSERT File X: X data_name "toto.txt", X data %(data)s',
                {"data": Binary(b"yo")},
            )[0][0]
            seid = cnx.create_entity("Salesterm", amount=0, described_by_test=feid).eid
            cnx.commit()

        # ensure a value that violate a constraint is properly detected
        with self.admin_access.web_request(rollbackfirst=True) as req:
            req.form = {
                "eid": [str(seid)],
                "__type:%s" % seid: "Salesterm",
                "_cw_entity_fields:%s" % seid: "amount-subject",
                "amount-subject:%s" % seid: "-10",
            }
            self.assertMultiLineEqual(
                """<script type="text/javascript">
 window.parent.handleFormValidationResponse('entityForm', null, null, [false, [%s, {"amount-subject": "value -10 must be >= 0"}], null], null);
</script>"""
                % seid,
                self.ctrl_publish(req, "validateform").decode("ascii"),
            )

        # ensure a value that comply a constraint is properly processed
        with self.admin_access.web_request(rollbackfirst=True) as req:
            req.form = {
                "eid": [str(seid)],
                "__type:%s" % seid: "Salesterm",
                "_cw_entity_fields:%s" % seid: "amount-subject",
                "amount-subject:%s" % seid: "20",
            }
            self.assertMultiLineEqual(
                f"""<script type="text/javascript">
 window.parent.handleFormValidationResponse('entityForm', null, null, [true, "{BASE_URL}view", null], null);
</script>""",
                self.ctrl_publish(req, "validateform").decode("ascii"),
            )
            self.assertEqual(
                20,
                req.execute("Any V WHERE X amount V, X eid %(eid)s", {"eid": seid})[0][
                    0
                ],
            )

        with self.admin_access.web_request(rollbackfirst=True) as req:
            req.form = {
                "eid": ["X"],
                "__type:X": "Salesterm",
                "_cw_entity_fields:X": "amount-subject,described_by_test-subject",
                "amount-subject:X": "0",
                "described_by_test-subject:X": str(feid),
            }

            # ensure a value that is modified in an operation on a modify
            # hook works as it should (see
            # https://www.cubicweb.org/ticket/2509729 )
            class MyOperation(Operation):
                def precommit_event(self):
                    self.entity.cw_set(amount=-10)

            class ValidationErrorInOpAfterHook(Hook):
                __regid__ = "valerror-op-after-hook"
                __select__ = Hook.__select__ & is_instance("Salesterm")
                events = ("after_add_entity",)

                def __call__(self):
                    MyOperation(self._cw, entity=self.entity)

            with self.temporary_appobjects(ValidationErrorInOpAfterHook):
                self.assertMultiLineEqual(
                    """<script type="text/javascript">
 window.parent.handleFormValidationResponse('entityForm', null, null, [false, ["X", {"amount-subject": "value -10 must be >= 0"}], null], null);
</script>""",
                    self.ctrl_publish(req, "validateform").decode("ascii"),
                )

            self.assertMultiLineEqual(
                f"""<script type="text/javascript">
 window.parent.handleFormValidationResponse('entityForm', null, null, [true, "{BASE_URL}view", null], null);
</script>""",
                self.ctrl_publish(req, "validateform").decode("ascii"),
            )

    def test_req_pending_insert(self):
        """make sure req's pending insertions are taken into account"""
        with self.admin_access.web_request() as req:
            tmpgroup = req.create_entity("CWGroup", name="test")
            user = req.user
            req.cnx.commit()
        with self.admin_access.web_request(**req_form(user)) as req:
            req.session.data["pending_insert"] = {(user.eid, "in_group", tmpgroup.eid)}
            self.expect_redirect_handle_request(req, "edit")
            usergroups = [
                gname
                for gname, in req.execute(
                    "Any N WHERE G name N, U in_group G, U eid %(u)s", {"u": user.eid}
                )
            ]
            self.assertCountEqual(usergroups, ["managers", "test"])
            self.assertEqual(get_pending_inserts(req), [])

    def test_req_pending_delete(self):
        """make sure req's pending deletions are taken into account"""
        with self.admin_access.web_request() as req:
            user = req.user
            groupeid = req.execute(
                'INSERT CWGroup G: G name "test", U in_group G WHERE U eid %(x)s',
                {"x": user.eid},
            )[0][0]
            usergroups = [
                gname
                for gname, in req.execute(
                    "Any N WHERE G name N, U in_group G, U eid %(u)s", {"u": user.eid}
                )
            ]
            # just make sure everything was set correctly
            self.assertCountEqual(usergroups, ["managers", "test"])
            req.cnx.commit()
            # now try to delete the relation
        with self.admin_access.web_request(**req_form(user)) as req:
            req.session.data["pending_delete"] = {(user.eid, "in_group", groupeid)}
            self.expect_redirect_handle_request(req, "edit")
            usergroups = [
                gname
                for gname, in req.execute(
                    "Any N WHERE G name N, U in_group G, U eid %(u)s", {"u": user.eid}
                )
            ]
            self.assertCountEqual(usergroups, ["managers"])
            self.assertEqual(get_pending_deletes(req), [])

    def test_redirect_apply_button(self):
        with self.admin_access.web_request() as req:
            redirectrql = rql_for_eid(4012)  # whatever
            req.form = {
                "eid": "A",
                "__maineid": "A",
                "__type:A": "BlogEntry",
                "_cw_entity_fields:A": "content-subject,title-subject",
                "content-subject:A": '"13:03:43"',
                "title-subject:A": "huuu",
                "__redirectrql": redirectrql,
                "__redirectvid": "primary",
                "__redirectparams": "toto=tutu&tata=titi",
                "__form_id": "edition",
                "__action_apply": "",
            }
            path, params = self.expect_redirect_handle_request(req, "edit")
            self.assertTrue(path.startswith("blogentry/"))
            eid = path.split("/")[1]
            self.assertEqual(params["vid"], "edition")
            self.assertNotEqual(int(eid), 4012)
            self.assertEqual(params["__redirectrql"], redirectrql)
            self.assertEqual(params["__redirectvid"], "primary")
            self.assertEqual(params["__redirectparams"], "toto=tutu&tata=titi")

    def test_redirect_ok_button(self):
        with self.admin_access.web_request() as req:
            redirectrql = rql_for_eid(4012)  # whatever
            req.form = {
                "eid": "A",
                "__maineid": "A",
                "__type:A": "BlogEntry",
                "_cw_entity_fields:A": "content-subject,title-subject",
                "content-subject:A": '"13:03:43"',
                "title-subject:A": "huuu",
                "__redirectrql": redirectrql,
                "__redirectvid": "primary",
                "__redirectparams": "toto=tutu&tata=titi",
                "__form_id": "edition",
            }
            path, params = self.expect_redirect_handle_request(req, "edit")
            self.assertEqual(path, "view")
            self.assertEqual(params["rql"], redirectrql)
            self.assertEqual(params["vid"], "primary")
            self.assertEqual(params["tata"], "titi")
            self.assertEqual(params["toto"], "tutu")

    def test_redirect_delete_button(self):
        with self.admin_access.web_request() as req:
            eid = req.create_entity("BlogEntry", title="hop", content="hop").eid
            req.form = {
                "eid": str(eid),
                "__type:%s" % eid: "BlogEntry",
                "__action_delete": "",
            }
            path, params = self.expect_redirect_handle_request(req, "edit")
            self.assertEqual(path, "blogentry")
            self.assertIn("_cwmsgid", params)
            eid = req.create_entity("EmailAddress", address="hop@logilab.fr").eid
            req.execute(
                "SET X use_email E WHERE E eid %(e)s, X eid %(x)s",
                {"x": req.user.eid, "e": eid},
            )
            req.cnx.commit()
            req.form = {
                "eid": str(eid),
                "__type:%s" % eid: "EmailAddress",
                "__action_delete": "",
            }
            path, params = self.expect_redirect_handle_request(req, "edit")
            self.assertEqual(path, "cwuser/admin")
            self.assertIn("_cwmsgid", params)
            eid1 = req.create_entity("BlogEntry", title="hop", content="hop").eid
            eid2 = req.create_entity("EmailAddress", address="hop@logilab.fr").eid
            req.form = {
                "eid": [str(eid1), str(eid2)],
                "__type:%s" % eid1: "BlogEntry",
                "__type:%s" % eid2: "EmailAddress",
                "__action_delete": "",
            }
            path, params = self.expect_redirect_handle_request(req, "edit")
            self.assertEqual(path, "view")
            self.assertIn("_cwmsgid", params)

    def test_simple_copy(self):
        with self.admin_access.web_request() as req:
            blog = req.create_entity("Blog", title="my-blog")
            blogentry = req.create_entity(
                "BlogEntry", title="entry1", content="content1", entry_of=blog
            )
            req.form = {
                "__maineid": "X",
                "eid": "X",
                "__cloned_eid:X": blogentry.eid,
                "__type:X": "BlogEntry",
                "_cw_entity_fields:X": "title-subject,content-subject",
                "title-subject:X": "entry1-copy",
                "content-subject:X": "content1",
            }
            self.expect_redirect_handle_request(req, "edit")
            blogentry2 = req.find("BlogEntry", title="entry1-copy").one()
            self.assertEqual(blogentry2.entry_of[0].eid, blog.eid)

    def test_skip_copy_for(self):
        with self.admin_access.web_request() as req:
            blog = req.create_entity("Blog", title="my-blog")
            blogentry = req.create_entity(
                "BlogEntry", title="entry1", content="content1", entry_of=blog
            )
            blogentry.__class__.cw_skip_copy_for = [("entry_of", "subject")]
            try:
                req.form = {
                    "__maineid": "X",
                    "eid": "X",
                    "__cloned_eid:X": blogentry.eid,
                    "__type:X": "BlogEntry",
                    "_cw_entity_fields:X": "title-subject,content-subject",
                    "title-subject:X": "entry1-copy",
                    "content-subject:X": "content1",
                }
                self.expect_redirect_handle_request(req, "edit")
                blogentry2 = req.find("BlogEntry", title="entry1-copy").one()
                # entry_of should not be copied
                self.assertEqual(len(blogentry2.entry_of), 0)
            finally:
                blogentry.__class__.cw_skip_copy_for = []

    def test_avoid_multiple_process_posted(self):
        # test that when some entity is being created and data include non-inlined relations, the
        # values for this relation are stored for later usage, without calling twice field's
        # process_form method, which may be unexpected for custom fields

        orig_process_posted = ff.RelationField.process_posted

        def count_process_posted(self, form):
            res = list(orig_process_posted(self, form))
            nb_process_posted_calls[0] += 1
            return res

        ff.RelationField.process_posted = count_process_posted

        try:
            with self.admin_access.web_request() as req:
                gueid = req.execute('CWGroup G WHERE G name "users"')[0][0]
                req.form = {
                    "eid": "X",
                    "__type:X": "CWUser",
                    "_cw_entity_fields:X": "login-subject,upassword-subject,in_group-subject",
                    "login-subject:X": "adim",
                    "upassword-subject:X": "toto",
                    "upassword-subject-confirm:X": "toto",
                    "in_group-subject:X": repr(gueid),
                }
                nb_process_posted_calls = [0]
                self.expect_redirect_handle_request(req, "edit")
                self.assertEqual(nb_process_posted_calls[0], 1)
                user = req.find("CWUser", login="adim").one()
                self.assertEqual({g.eid for g in user.in_group}, {gueid})
                req.form = {
                    "eid": ["X", "Y"],
                    "__type:X": "CWUser",
                    "_cw_entity_fields:X": "login-subject,upassword-subject,in_group-subject",
                    "login-subject:X": "dlax",
                    "upassword-subject:X": "toto",
                    "upassword-subject-confirm:X": "toto",
                    "in_group-subject:X": repr(gueid),
                    "__type:Y": "EmailAddress",
                    "_cw_entity_fields:Y": "address-subject,use_email-object",
                    "address-subject:Y": "dlax@cw.org",
                    "use_email-object:Y": "X",
                }
                nb_process_posted_calls = [0]
                self.expect_redirect_handle_request(req, "edit")
                self.assertEqual(
                    nb_process_posted_calls[0], 3
                )  # 3 = 1 (in_group) + 2 (use_email)
                user = req.find("CWUser", login="dlax").one()
                self.assertEqual({e.address for e in user.use_email}, {"dlax@cw.org"})

        finally:
            ff.RelationField.process_posted = orig_process_posted

    def test_nonregr_eetype_etype_editing(self):
        """non-regression test checking that a manager user can edit a CWEType entity"""
        with self.admin_access.web_request() as req:
            groupeids = sorted(
                eid
                for eid, in req.execute(
                    "CWGroup G " 'WHERE G name in ("managers", "users")'
                )
            )
            groups = [str(eid) for eid in groupeids]
            cwetypeeid = req.execute('CWEType X WHERE X name "CWEType"')[0][0]
            basegroups = [
                str(eid)
                for eid, in req.execute(
                    "CWGroup G " "WHERE X read_permission G, X eid %(x)s",
                    {"x": cwetypeeid},
                )
            ]
            cwetypeeid = str(cwetypeeid)
            req.form = {
                "eid": cwetypeeid,
                "__type:" + cwetypeeid: "CWEType",
                "_cw_entity_fields:"
                + cwetypeeid: "name-subject,final-subject,description-subject,read_permission-subject",
                "name-subject:" + cwetypeeid: "CWEType",
                "final-subject:" + cwetypeeid: "",
                "description-subject:" + cwetypeeid: "users group",
                "read_permission-subject:" + cwetypeeid: groups,
            }
            try:
                self.expect_redirect_handle_request(req, "edit")
                e = req.execute(
                    "Any X WHERE X eid %(x)s", {"x": cwetypeeid}
                ).get_entity(0, 0)
                self.assertEqual(e.name, "CWEType")
                self.assertEqual(sorted(g.eid for g in e.read_permission), groupeids)
            finally:
                # restore
                req.execute(
                    'SET X read_permission Y WHERE X name "CWEType", '
                    "Y eid IN (%s), NOT X read_permission Y" % (",".join(basegroups))
                )
                req.cnx.commit()

    def test_nonregr_strange_text_input(self):
        """non-regression test checking text input containing "13:03:43"

        this seems to be postgres (tsearch?) specific
        """
        with self.admin_access.web_request() as req:
            req.form = {
                "eid": "A",
                "__maineid": "A",
                "__type:A": "BlogEntry",
                "_cw_entity_fields:A": "title-subject,content-subject",
                "title-subject:A": '"13:03:40"',
                "content-subject:A": '"13:03:43"',
            }
            path, _params = self.expect_redirect_handle_request(req, "edit")
            self.assertTrue(path.startswith("blogentry/"))
            eid = path.split("/")[1]
            e = req.execute(
                "Any C, T WHERE C eid %(x)s, C content T", {"x": eid}
            ).get_entity(0, 0)
            self.assertEqual(e.title, '"13:03:40"')
            self.assertEqual(e.content, '"13:03:43"')

    def test_nonregr_multiple_empty_email_addr(self):
        with self.admin_access.web_request() as req:
            gueid = req.execute('CWGroup G WHERE G name "users"')[0][0]
            req.form = {
                "eid": ["X", "Y"],
                "__type:X": "CWUser",
                "_cw_entity_fields:X": "login-subject,upassword-subject,in_group-subject",
                "login-subject:X": "adim",
                "upassword-subject:X": "toto",
                "upassword-subject-confirm:X": "toto",
                "in_group-subject:X": repr(gueid),
                "__type:Y": "EmailAddress",
                "_cw_entity_fields:Y": "address-subject,alias-subject,use_email-object",
                "address-subject:Y": "",
                "alias-subject:Y": "",
                "use_email-object:Y": "X",
            }
            with self.assertRaises(ValidationError) as cm:
                self.ctrl_publish(req)
            self.assertEqual(cm.exception.errors, {"address-subject": "required field"})

    def test_nonregr_copy(self):
        with self.admin_access.web_request() as req:
            user = req.user
            req.form = {
                "__maineid": "X",
                "eid": "X",
                "__cloned_eid:X": user.eid,
                "__type:X": "CWUser",
                "_cw_entity_fields:X": "login-subject,upassword-subject",
                "login-subject:X": "toto",
                "upassword-subject:X": "toto",
                "upassword-subject-confirm:X": "toto",
            }
            path, _params = self.expect_redirect_handle_request(req, "edit")
            self.assertEqual(path, "cwuser/toto")
            e = req.execute('Any X WHERE X is CWUser, X login "toto"').get_entity(0, 0)
            self.assertEqual(e.login, "toto")
            self.assertEqual(e.in_group[0].name, "managers")

    def test_nonregr_rollback_on_validation_error(self):
        with self.admin_access.web_request(url="edit") as req:
            p = self.create_user(req, "doe")
            # do not try to skip 'primary_email' for this test
            old_skips = p.__class__.cw_skip_copy_for
            p.__class__.cw_skip_copy_for = ()
            try:
                e = req.create_entity("EmailAddress", address="doe@doe.com")
                req.execute(
                    "SET P use_email E, P primary_email E WHERE P eid %(p)s, E eid %(e)s",
                    {"p": p.eid, "e": e.eid},
                )
                req.form = {
                    "eid": "X",
                    "__cloned_eid:X": p.eid,
                    "__type:X": "CWUser",
                    "_cw_entity_fields:X": "login-subject,surname-subject",
                    "login-subject": "dodo",
                    "surname-subject:X": "Boom",
                    "__errorurl": "whatever but required",
                }
                # try to emulate what really happens in the web application
                # 1/ validate form => EditController.publish raises a ValidationError
                #    which fires a Redirect
                # 2/ When re-publishing the copy form, the publisher implicitly commits
                try:
                    self.app_handle_request(req)
                except Redirect:
                    req.form["rql"] = "Any X WHERE X eid %s" % p.eid
                    req.form["vid"] = "copy"
                    self.app_handle_request(req, "view")
                rset = req.execute('CWUser P WHERE P surname "Boom"')
                self.assertEqual(len(rset), 0)
            finally:
                p.__class__.cw_skip_copy_for = old_skips

    def test_regr_inlined_forms(self):
        with self.admin_access.web_request() as req:
            self.schema["described_by_test"].inlined = False
            try:
                req.data["eidmap"] = {}
                req.data["pending_others"] = set()
                req.data["pending_inlined"] = {}
                req.form = {
                    "eid": ["X", "Y"],
                    "__maineid": "X",
                    "__type:X": "Salesterm",
                    "_cw_entity_fields:X": "described_by_test-subject",
                    "described_by_test-subject:X": "Y",
                    "__type:Y": "File",
                    "_cw_entity_fields:Y": "data-subject",
                    "data-subject:Y": ("coucou.txt", Binary(b"coucou")),
                }
                values_by_eid = {
                    eid: req.extract_entity_params(eid, minparams=2)
                    for eid in req.edited_eids()
                }
                editctrl = self.vreg["controllers"].select("edit", req)
                # don't call publish to enforce select order
                editctrl.errors = []
                editctrl._to_create = {}
                editctrl.edit_entity(
                    values_by_eid["X"]
                )  # #3064653 raise ValidationError
                editctrl.edit_entity(values_by_eid["Y"])
            finally:
                self.schema["described_by_test"].inlined = False


class ReportBugControllerTC(WebCWTC):
    def test_usable_by_guest(self):
        with self.new_access("anon").web_request() as req:
            self.assertRaises(
                NoSelectableObject, self.vreg["controllers"].select, "reportbug", req
            )
        with self.new_access("anon").web_request(description="hop") as req:
            self.vreg["controllers"].select("reportbug", req)


class AjaxControllerTC(WebCWTC):
    tested_controller = "ajax"

    def ctrl(self, req):
        return self.vreg["controllers"].select(self.tested_controller, req)

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            self.pytag = cnx.create_entity("Tag", name="python")
            self.cubicwebtag = cnx.create_entity("Tag", name="cubicweb")
            self.john = self.create_user(cnx, "John")
            cnx.commit()

    # tests ##################################################################
    def test_simple_exec(self):
        with self.admin_access.web_request(
            rql='CWUser P WHERE P login "John"', pageid="123", fname="view"
        ) as req:
            ctrl = self.ctrl(req)
            rset = self.john.as_rset()
            rset.req = req
            source = ctrl.publish()
            self.assertTrue(source.startswith(b"<div>"))

    #     def test_json_exec(self):
    #         rql = 'Any T,N WHERE T is Tag, T name N'
    #         ctrl = self.ctrl(self.request(mode='json', rql=rql, pageid='123'))
    #         self.assertEqual(ctrl.publish(),
    #                           json_dumps(self.execute(rql).rows))

    def test_remote_add_existing_tag(self):
        with self.remote_calling("tag_entity", self.john.eid, ["python"]) as (_, req):
            self.assertCountEqual(
                [tname for tname, in req.execute("Any N WHERE T is Tag, T name N")],
                ["python", "cubicweb"],
            )
            self.assertEqual(
                req.execute("Any N WHERE T tags P, P is CWUser, T name N").rows,
                [["python"]],
            )

    def test_remote_add_new_tag(self):
        with self.remote_calling("tag_entity", self.john.eid, ["javascript"]) as (
            _,
            req,
        ):
            self.assertCountEqual(
                [tname for tname, in req.execute("Any N WHERE T is Tag, T name N")],
                ["python", "cubicweb", "javascript"],
            )
            self.assertEqual(
                req.execute("Any N WHERE T tags P, P is CWUser, T name N").rows,
                [["javascript"]],
            )

    def test_maydel_perms(self):
        """Check that AjaxEditRelationCtxComponent calls rdef.check with a
        sufficient context"""
        with self.remote_calling("tag_entity", self.john.eid, ["python"]) as (_, req):
            req.cnx.commit()
        with self.temporary_permissions(
            (
                self.schema["tags"].relation_definitions["Tag", "CWUser"],
                {"delete": (RRQLExpression("S owned_by U"),)},
            )
        ):
            with self.admin_access.web_request(
                rql='CWUser P WHERE P login "John"',
                pageid="123",
                fname="view",
                session=req.session,
            ) as req:
                ctrl = self.ctrl(req)
                rset = self.john.as_rset()
                rset.req = req
                source = ctrl.publish()
                # maydel jscall
                self.assertIn(b"ajaxBoxRemoveLinkedEntity", source)

    def test_pending_insertion(self):
        with self.remote_calling("add_pending_inserts", [["12", "tags", "13"]]) as (
            _,
            req,
        ):
            deletes = get_pending_deletes(req)
            self.assertEqual(deletes, [])
            inserts = get_pending_inserts(req)
            self.assertEqual(inserts, ["12:tags:13"])
        with self.remote_calling(
            "add_pending_inserts", [["12", "tags", "14"]], session=req.session
        ) as (_, req):
            deletes = get_pending_deletes(req)
            self.assertEqual(deletes, [])
            inserts = get_pending_inserts(req)
            self.assertCountEqual(inserts, ["12:tags:13", "12:tags:14"])
            inserts = get_pending_inserts(req, 12)
            self.assertCountEqual(inserts, ["12:tags:13", "12:tags:14"])
            inserts = get_pending_inserts(req, 13)
            self.assertEqual(inserts, ["12:tags:13"])
            inserts = get_pending_inserts(req, 14)
            self.assertEqual(inserts, ["12:tags:14"])
            req.remove_pending_operations()

    def test_pending_deletion(self):
        with self.remote_calling("add_pending_delete", ["12", "tags", "13"]) as (
            _,
            req,
        ):
            inserts = get_pending_inserts(req)
            self.assertEqual(inserts, [])
            deletes = get_pending_deletes(req)
            self.assertEqual(deletes, ["12:tags:13"])
        with self.remote_calling(
            "add_pending_delete", ["12", "tags", "14"], session=req.session
        ) as (_, req):
            inserts = get_pending_inserts(req)
            self.assertEqual(inserts, [])
            deletes = get_pending_deletes(req)
            self.assertCountEqual(deletes, ["12:tags:13", "12:tags:14"])
            deletes = get_pending_deletes(req, 12)
            self.assertCountEqual(deletes, ["12:tags:13", "12:tags:14"])
            deletes = get_pending_deletes(req, 13)
            self.assertEqual(deletes, ["12:tags:13"])
            deletes = get_pending_deletes(req, 14)
            self.assertEqual(deletes, ["12:tags:14"])
            req.remove_pending_operations()

    def test_remove_pending_operations(self):
        with self.remote_calling("add_pending_delete", ["12", "tags", "13"]) as (
            _,
            req,
        ):
            pass
        with self.remote_calling(
            "add_pending_inserts", [["12", "tags", "14"]], session=req.session
        ) as (_, req):
            inserts = get_pending_inserts(req)
            self.assertEqual(inserts, ["12:tags:14"])
            deletes = get_pending_deletes(req)
            self.assertEqual(deletes, ["12:tags:13"])
            req.remove_pending_operations()
            self.assertEqual(get_pending_deletes(req), [])
            self.assertEqual(get_pending_inserts(req), [])

    def test_add_inserts(self):
        with self.remote_calling(
            "add_pending_inserts", [("12", "tags", "13"), ("12", "tags", "14")]
        ) as (_, req):
            inserts = get_pending_inserts(req)
            self.assertCountEqual(inserts, ["12:tags:13", "12:tags:14"])
            req.remove_pending_operations()

    # silly tests
    def test_external_resource(self):
        with self.remote_calling("external_resource", "RSS_LOGO") as (res, _):
            self.assertEqual(
                json_dumps(self.config.uiprops["RSS_LOGO"]).encode("ascii"), res
            )

    def test_i18n(self):
        with self.remote_calling("i18n", ["bimboom"]) as (res, _):
            self.assertEqual(json_dumps(["bimboom"]).encode("ascii"), res)

    def test_format_date(self):
        with self.remote_calling("format_date", "2007-01-01 12:00:00") as (res, _):
            self.assertEqual(json_dumps("2007/01/01").encode("ascii"), res)

    def test_ajaxfunc_noparameter(self):
        @ajaxfunc
        def foo(self, x, y):
            return "hello"

        self.assertEqual(foo(object, 1, 2), "hello")
        appobject = foo.__appobject__
        self.assertTrue(issubclass(appobject, AjaxFunction))
        self.assertEqual(appobject.__regid__, "foo")
        self.assertEqual(appobject.check_pageid, False)
        self.assertEqual(appobject.output_type, None)
        with self.admin_access.web_request() as req:
            f = appobject(req)
            self.assertEqual(f(12, 13), "hello")

    def test_ajaxfunc_checkpageid(self):
        @ajaxfunc(check_pageid=True)
        def foo(self, x, y):
            return "hello"

        self.assertEqual(foo(object, 1, 2), "hello")
        appobject = foo.__appobject__
        self.assertTrue(issubclass(appobject, AjaxFunction))
        self.assertEqual(appobject.__regid__, "foo")
        self.assertEqual(appobject.check_pageid, True)
        self.assertEqual(appobject.output_type, None)
        # no pageid
        with self.admin_access.web_request() as req:
            f = appobject(req)
            self.assertRaises(RemoteCallFailed, f, 12, 13)

    def test_ajaxfunc_json(self):
        @ajaxfunc(output_type="json")
        def foo(self, x, y):
            return x + y

        self.assertEqual(foo(object, 1, 2), 3)
        appobject = foo.__appobject__
        self.assertTrue(issubclass(appobject, AjaxFunction))
        self.assertEqual(appobject.__regid__, "foo")
        self.assertEqual(appobject.check_pageid, False)
        self.assertEqual(appobject.output_type, "json")
        # no pageid
        with self.admin_access.web_request() as req:
            f = appobject(req)
            self.assertEqual(f(12, 13), "25")

    def test_badrequest(self):
        with self.assertRaises(RemoteCallFailed) as cm:
            with self.remote_calling("foo"):
                pass
        self.assertEqual(cm.exception.status, 400)
        self.assertEqual(cm.exception.reason, "no foo method")


class UndoControllerTC(WebCWTC):
    def setUp(self):
        # Force undo feature to be turned on
        Connection.undo_actions = property(lambda self: True, lambda self, v: None)
        super().setUp()

    def tearDown(self):
        super().tearDown()
        del Connection.undo_actions

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            self.toto = self.create_user(
                cnx, "toto", password="toto", groups=("users",), commit=False
            )
            self.txuuid_toto = cnx.commit()
            self.toto_email = cnx.create_entity(
                "EmailAddress", address="toto@logilab.org", reverse_use_email=self.toto
            )
            self.txuuid_toto_email = cnx.commit()

    def test_no_such_transaction(self):
        with self.admin_access.web_request() as req:
            txuuid = "12345acbd"
            req.form["txuuid"] = txuuid
            controller = self.vreg["controllers"].select("undo", req)
            with self.assertRaises(tx.NoSuchTransaction) as cm:
                controller.publish(rset=None)
            self.assertEqual(cm.exception.txuuid, txuuid)

    def assertURLPath(self, url, expected_path, expected_params=None):
        """This assert that the path part of `url` matches  expected path

        TODO : implement assertion on the expected_params too
        """
        with self.admin_access.web_request() as req:
            scheme, netloc, path, query, fragment = urlsplit(url)
            parse_qs(query)
            expected_url = urljoin(req.base_url(), expected_path)
            self.assertEqual(
                urlunsplit((scheme, netloc, path, None, None)), expected_url
            )

    def test_redirect_redirectpath(self):
        "Check that the potential __redirectpath is honored"
        with self.admin_access.web_request() as req:
            txuuid = self.txuuid_toto_email
            req.form["txuuid"] = txuuid
            rpath = "toto"
            req.form["__redirectpath"] = rpath
            controller = self.vreg["controllers"].select("undo", req)
            with self.assertRaises(Redirect) as cm:
                controller.publish(rset=None)
            self.assertURLPath(cm.exception.location, rpath)


class LoginControllerTC(WebCWTC):
    def test_login_with_dest(self):
        with self.admin_access.web_request() as req:
            req.form = {"postlogin_path": "elephants/babar"}
            with self.assertRaises(Redirect) as cm:
                self.ctrl_publish(req, ctrl="login")
            self.assertEqual(req.build_url("elephants/babar"), cm.exception.location)

    def test_login_no_dest(self):
        with self.admin_access.web_request() as req:
            with self.assertRaises(Redirect) as cm:
                self.ctrl_publish(req, ctrl="login")
            self.assertEqual(req.base_url(), cm.exception.location)


class LoginControllerHTTPTC(PyramidWebCWTC):
    anonymous_allowed = True
    # this TC depends on the auth mode being 'cookie' and not 'http'
    # (the former being the default, so everything works)

    def test_http_error_codes_auth_fail(self):
        response = self.webapp.login(user="toto", password="pouetA", status=403)
        self.assertEqual(response.status_code, 403)

    def test_http_error_codes_auth_succeed(self):
        self.webapp.login(user=self.admlogin, password=self.admpassword)


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Views, forms, actions... for the CubicWeb web client"""

from logilab.mtconverter import xml_escape
from rql import nodes

from cubicweb_web.views._vid_by_mimetype import VID_BY_MIMETYPE


def need_table_view(rset, schema):
    """return True if we think that a table view is more appropriate than a
    list or primary view to display the given result set
    """
    rqlst = rset.syntax_tree()
    if len(rqlst.children) > 1:
        # UNION query, use a table
        return True
    selected = rqlst.children[0].selection
    try:
        mainvar = selected[0]
    except AttributeError:
        # not a variable ref, using table view is probably a good option
        return True
    if not (
        isinstance(mainvar, nodes.VariableRef)
        or (isinstance(mainvar, nodes.Constant) and mainvar.uid)
    ):
        return True
    for i, etype in enumerate(rset.description[0][1:]):
        # etype may be None on outer join
        if etype is None:
            return True
        # check the selected index node is a VariableRef (else we
        # won't detect aggregate function
        if not isinstance(selected[i + 1], nodes.VariableRef):
            return True
        # if this is not a final entity
        if not schema.entity_schema_for(etype).final:
            return True
        # if this is a final entity not linked to the main variable
        var = selected[i + 1].variable
        for vref in var.references():
            rel = vref.relation()
            if rel is None:
                continue
            if mainvar.is_equivalent(rel.children[0]):
                break
        else:
            return True
    return False


def vid_from_rset(req, rset, schema, check_table=True):
    """given a result set, return a view id"""
    if rset is None:
        return "index"
    for mimetype in req.parse_accept_header("Accept"):
        if mimetype in VID_BY_MIMETYPE:
            return VID_BY_MIMETYPE[mimetype]
    nb_rows = len(rset)
    # empty resultset
    if nb_rows == 0:
        return "noresult"
    # entity result set
    if not schema.entity_schema_for(rset.description[0][0]).final:
        if check_table and need_table_view(rset, schema):
            return "table"
        if nb_rows == 1:
            if req.search_state[0] == "normal":
                return "primary"
            return "outofcontext-search"
        if len(rset.column_types(0)) == 1:
            return "sameetypelist"
        return "list"
    return "table"


def linksearch_select_url(req, rset):
    """when searching an entity to create a relation, return a URL to select
    entities in the given rset
    """
    req.add_js(("cubicweb.ajax.js", "cubicweb.edition.js"))
    target, eid, r_type, searchedtype = req.search_state[1]
    if target == "subject":
        id_fmt = f"{eid}:{r_type}:%s"
    else:
        id_fmt = f"%s:{r_type}:{eid}"
    triplets = "-".join(id_fmt % row[0] for row in rset.rows)
    return f"javascript: selectForAssociation('{triplets}', '{eid}');"


def add_etype_button(req, etype, csscls="addButton right", **urlkwargs):
    vreg = req.vreg
    eschema = vreg.schema.entity_schema_for(etype)
    if eschema.has_perm(req, "add"):
        url = vreg["etypes"].etype_class(etype).cw_create_url(req, **urlkwargs)
        return '<a href="{}" class="{}">{}</a>'.format(
            xml_escape(url),
            csscls,
            req.__("New %s" % etype),
        )
    return ""



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
from logilab.common.testlib import unittest_main

from cubicweb_web.devtools.testlib import WebCWTC


class PyViewsCWTC(WebCWTC):
    def test_pyvaltable(self):
        with self.admin_access.web_request() as req:
            view = self.vreg["views"].select(
                "pyvaltable", req, pyvalue=[[1, "a"], [2, "b"]]
            )
            content = view.render(pyvalue=[[1, "a"], [2, "b"]], headers=["num", "char"])
            self.assertEqual(
                content.strip(),
                """<table class="listing"><tbody>\
<tr class="even" onmouseout="$(this).removeClass(&quot;highlighted&quot;)" onmouseover="$(this).addClass(&quot;highlighted&quot;);"><td >1</td><td >a</td></tr>
<tr class="odd" onmouseout="$(this).removeClass(&quot;highlighted&quot;)" onmouseover="$(this).addClass(&quot;highlighted&quot;);"><td >2</td><td >b</td></tr>
</tbody></table>""",
            )

    def test_pyvallist(self):
        with self.admin_access.web_request() as req:
            view = self.vreg["views"].select("pyvallist", req, pyvalue=[1, "a"])
            content = view.render(pyvalue=[1, "a"])
            self.assertEqual(
                content.strip(),
                """<ul>
<li>1</li>
<li>a</li>
</ul>""",
            )


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
Renderers
---------

.. Note::
   Form renderers are responsible to layout a form to HTML.

Here are the base renderers available:

.. autoclass:: cubicweb_web.views.formrenderers.FormRenderer
.. autoclass:: cubicweb_web.views.formrenderers.HTableFormRenderer
.. autoclass:: cubicweb_web.views.formrenderers.EntityCompositeFormRenderer
.. autoclass:: cubicweb_web.views.formrenderers.EntityFormRenderer
.. autoclass:: cubicweb_web.views.formrenderers.EntityInlinedFormRenderer

"""


from logilab.common.registry import yes
from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb import uilib
from cubicweb.appobject import AppObject
from cubicweb.predicates import is_instance
from cubicweb.uilib import eid_param
from cubicweb.utils import json_dumps
from cubicweb_web import formwidgets as fwdgs
from cubicweb_web import tags


def checkbox(name, value, attrs="", checked=None):
    if checked is None:
        checked = value
    checked = checked and 'checked="checked"' or ""
    return '<input type="checkbox" name="{}" value="{}" {} {} />'.format(
        name,
        value,
        checked,
        attrs,
    )


def field_label(form, field):
    if callable(field.label):
        return field.label(form, field)
    # XXX with 3.6 we can now properly rely on 'if field.role is not None' and
    # stop having a tuple for label
    if isinstance(field.label, tuple):  # i.e. needs contextual translation
        return form._cw.pgettext(*field.label)
    return form._cw._(field.label)


class FormRenderer(AppObject):
    """This is the 'default' renderer, displaying fields in a two columns table:

    +--------------+--------------+
    | field1 label | field1 input |
    +--------------+--------------+
    | field2 label | field2 input |
    +--------------+--------------+

    +---------+
    | buttons |
    +---------+
    """

    __registry__ = "formrenderers"
    __regid__ = "default"

    _options = (
        "display_label",
        "display_help",
        "display_progress_div",
        "table_class",
        "button_bar_class",
        # add entity since it may be given to select the renderer
        "entity",
    )
    display_label = True
    display_help = True
    display_progress_div = True
    table_class = "attributeForm"
    button_bar_class = "formButtonBar"

    def __init__(self, req=None, rset=None, row=None, col=None, **kwargs):
        super().__init__(req, rset=rset, row=row, col=col)
        if self._set_options(kwargs):
            raise ValueError("unconsumed arguments %s" % kwargs)

    def _set_options(self, kwargs):
        for key in self._options:
            try:
                setattr(self, key, kwargs.pop(key))
            except KeyError:
                continue
        return kwargs

    # renderer interface ######################################################

    def render(self, w, form, values):
        self._set_options(values)
        form.add_media()
        data = []
        _w = data.append
        _w(self.open_form(form, values))
        self.render_content(_w, form, values)
        _w(self.close_form(form, values))
        errormsg = self.error_message(form)
        if errormsg:
            data.insert(0, errormsg)
        # NOTE: we call str because `tag` objects may be found within data
        #       e.g. from the cwtags library
        w("".join(str(x) for x in data))

    def render_content(self, w, form, values):
        if self.display_progress_div:
            w('<div id="progress">%s</div>' % self._cw._("validating..."))
        w("\n<fieldset>\n")
        self.render_fields(w, form, values)
        self.render_buttons(w, form)
        w("\n</fieldset>\n")

    def render_label(self, form, field):
        if field.label is None:
            return ""
        label = field_label(form, field)
        attrs = {"for": field.dom_id(form)}
        if field.required:
            attrs["class"] = "required"
        return tags.label(label, **attrs)

    def render_help(self, form, field):
        help = []
        descr = field.help
        if callable(descr):
            descr = descr(form, field)
        if descr:
            help.append('<div class="helper">%s</div>' % self._cw._(descr))
        example = field.example_format(self._cw)
        if example:
            help.append(
                '<div class="helper">(%s: %s)</div>'
                % (self._cw._("sample format"), example)
            )
        return "&#160;".join(help)

    # specific methods (mostly to ease overriding) #############################

    def error_message(self, form):
        """return formatted error message

        This method should be called once inlined field errors has been consumed
        """
        req = self._cw
        errex = form.form_valerror
        # get extra errors
        if errex is not None:
            errormsg = req._("please correct the following errors:")
            errors = form.remaining_errors()
            if errors:
                if len(errors) > 1:
                    templstr = "<li>%s</li>\n"
                else:
                    templstr = "&#160;%s\n"
                for field, err in errors:
                    if not field:
                        errormsg += templstr % err
                    else:
                        errormsg += templstr % "%s: %s" % (req._(field), err)
                if len(errors) > 1:
                    errormsg = "<ul>%s</ul>" % errormsg
            return '<div class="errorMessage">%s</div>' % errormsg
        return ""

    def open_form(self, form, values, **attrs):
        if form.needs_multipart:
            enctype = "multipart/form-data"
        else:
            enctype = "application/x-www-form-urlencoded"
        attrs.setdefault("enctype", enctype)
        attrs.setdefault("method", "post")
        attrs.setdefault("action", form.form_action() or "#")
        if form.domid:
            attrs.setdefault("id", form.domid)
        if form.onsubmit:
            attrs.setdefault("onsubmit", form.onsubmit)
        if form.cssstyle:
            attrs.setdefault("style", form.cssstyle)
        if form.cssclass:
            attrs.setdefault("class", form.cssclass)
        if form.cwtarget:
            attrs.setdefault("target", form.cwtarget)
        if not form.autocomplete:
            attrs.setdefault("autocomplete", "off")
        return "<form %s>" % uilib.sgml_attributes(attrs)

    def close_form(self, form, values):
        """seems dumb but important for consistency w/ close form, and necessary
        for form renderers overriding open_form to use something else or more than
        and <form>
        """
        out = "</form>"
        if form.cwtarget:
            attrs = {
                "name": form.cwtarget,
                "id": form.cwtarget,
                "width": "0px",
                "height": "0px",
                "src": "javascript: void(0);",
            }
            out = ("<iframe %s></iframe>\n" % uilib.sgml_attributes(attrs)) + out
        return out

    def render_fields(self, w, form, values):
        fields = self._render_hidden_fields(w, form)
        if fields:
            self._render_fields(fields, w, form)
        self.render_child_forms(w, form, values)

    def render_child_forms(self, w, form, values):
        # render
        for childform in getattr(form, "forms", []):
            self.render_fields(w, childform, values)

    def _render_hidden_fields(self, w, form):
        fields = form.fields[:]
        for field in form.fields:
            if not field.is_visible():
                w(field.render(form, self))
                w("\n")
                fields.remove(field)
        return fields

    def _render_fields(self, fields, w, form):
        byfieldset = {}
        for field in fields:
            byfieldset.setdefault(field.fieldset, []).append(field)
        if form.fieldsets_in_order:
            fieldsets = form.fieldsets_in_order
        else:
            fieldsets = byfieldset
        for fieldset in list(fieldsets):
            try:
                fields = byfieldset.pop(fieldset)
            except KeyError:
                self.warning("no such fieldset: %s (%s)", fieldset, form)
                continue
            w("<fieldset>\n")
            if fieldset:
                w("<legend>%s</legend>" % self._cw.__(fieldset))
            w('<table class="%s">\n' % self.table_class)
            for field in fields:
                w(f'<tr class="{field.name}_{field.role}_row">\n')
                if self.display_label and field.label is not None:
                    w('<th class="labelCol">%s</th>\n' % self.render_label(form, field))
                w("<td")
                if field.label is None:
                    w(' colspan="2"')
                error = form.field_error(field)
                if error:
                    w(' class="error"')
                w(">\n")
                w(field.render(form, self))
                w("\n")
                if error:
                    self.render_error(w, error)
                if self.display_help:
                    w(self.render_help(form, field))
                w("</td></tr>\n")
            w("</table></fieldset>\n")
        if byfieldset:
            self.warning("unused fieldsets: %s", ", ".join(byfieldset))

    def render_buttons(self, w, form):
        if not form.form_buttons:
            return
        w('<table class="%s">\n<tr>\n' % self.button_bar_class)
        for button in form.form_buttons:
            w("<td>%s</td>\n" % button.render(form))
        w("</tr></table>")

    def render_error(self, w, err):
        """return validation error for widget's field, if any"""
        w('<span class="errorMsg">%s</span>' % err)


class BaseFormRenderer(FormRenderer):
    """use form_renderer_id = 'base' if you want base FormRenderer layout even
    when selected for an entity
    """

    __regid__ = "base"


class HTableFormRenderer(FormRenderer):
    """The 'htable' form renderer display fields horizontally in a table:

    +--------------+--------------+---------+
    | field1 label | field2 label |         |
    +--------------+--------------+---------+
    | field1 input | field2 input | buttons |
    +--------------+--------------+---------+
    """

    __regid__ = "htable"

    display_help = False

    def _render_fields(self, fields, w, form):
        w('<table border="0" class="htableForm">')
        if self.display_label:
            w("<tr>")
            for field in fields:
                w('<th class="labelCol">%s</th>' % self.render_label(form, field))
                if self.display_help:
                    w(self.render_help(form, field))
            # empty slot for buttons
            w('<th class="labelCol">&#160;</th>')
            w("</tr>")
        w("<tr>")
        for field in fields:
            error = form.field_error(field)
            if error:
                w('<td class="error">')
                self.render_error(w, error)
            else:
                w("<td>")
            w(field.render(form, self))
            w("</td>")
        w("<td>")
        for button in form.form_buttons:
            w(button.render(form))
        w("</td>")
        w("</tr>")
        w("</table>")

    def render_buttons(self, w, form):
        pass


class OneRowTableFormRenderer(FormRenderer):
    """The 'htable' form renderer display fields horizontally in a table:

    +--------------+--------------+--------------+--------------+---------+
    | field1 label | field1 input | field2 label | field2 input | buttons |
    +--------------+--------------+--------------+--------------+---------+
    """

    __regid__ = "onerowtable"

    display_help = False

    def _render_fields(self, fields, w, form):
        w('<table border="0" class="oneRowTableForm">')
        w("<tr>")
        for field in fields:
            if self.display_label:
                w('<th class="labelCol">%s</th>' % self.render_label(form, field))
            if self.display_help:
                w(self.render_help(form, field))
            error = form.field_error(field)
            if error:
                w('<td class="error">')
                self.render_error(w, error)
            else:
                w("<td>")
            w(field.render(form, self))
            w("</td>")
        w("<td>")
        for button in form.form_buttons:
            w(button.render(form))
        w("</td>")
        w("</tr>")
        w("</table>")

    def render_buttons(self, w, form):
        pass


class EntityCompositeFormRenderer(FormRenderer):
    """This is a specific renderer for the multiple entities edition form
    ('muledit').

    Each entity form will be displayed in row off a table, with a check box for
    each entities to indicate which ones are edited. Those checkboxes should be
    automatically updated when something is edited.
    """

    __regid__ = "composite"

    _main_display_fields = None

    def render_fields(self, w, form, values):
        if form.parent_form is None:
            w('<table class="listing">')
            # get fields from the first subform with something to display (we
            # may have subforms with nothing editable that will simply be
            # skipped later)
            for subform in form.forms:
                subfields = [field for field in subform.fields if field.is_visible()]
                if subfields:
                    break
            if subfields:
                # main form, display table headers
                w('<tr class="header">')
                w(
                    '<th align="left">%s</th>'
                    % tags.input(
                        type="checkbox",
                        title=self._cw._("toggle check boxes"),
                        onclick="setCheckboxesState('eid', null, this.checked)",
                    )
                )
                for field in subfields:
                    w("<th>%s</th>" % field_label(form, field))
                w("</tr>")
        super().render_fields(w, form, values)
        if form.parent_form is None:
            w("</table>")
            if self._main_display_fields:
                super()._render_fields(self._main_display_fields, w, form)

    def _render_fields(self, fields, w, form):
        if form.parent_form is not None:
            entity = form.edited_entity
            values = form.form_previous_values
            qeid = eid_param("eid", entity.eid)
            cbsetstate = "setCheckboxesState('eid', %s, 'checked')" % xml_escape(
                json_dumps(entity.eid)
            )
            w('<tr class="%s">' % (entity.cw_row % 2 and "even" or "odd"))
            # XXX turn this into a widget used on the eid field
            w("<td>%s</td>" % checkbox("eid", entity.eid, checked=qeid in values))
            for field in fields:
                error = form.field_error(field)
                if error:
                    w('<td class="error">')
                    self.render_error(w, error)
                else:
                    w("<td>")
                if isinstance(
                    field.widget, (fwdgs.Select, fwdgs.CheckBox, fwdgs.Radio)
                ):
                    field.widget.attrs["onchange"] = cbsetstate
                elif isinstance(field.widget, fwdgs.Input):
                    field.widget.attrs["onkeypress"] = cbsetstate
                # XXX else
                w("<div>%s</div>" % field.render(form, self))
                w("</td>\n")
            w("</tr>")
        else:
            self._main_display_fields = fields


class EntityFormRenderer(BaseFormRenderer):
    """This is the 'default' renderer for entity's form.

    You can still use form_renderer_id = 'base' if you want base FormRenderer
    layout even when selected for an entity.
    """

    __regid__ = "default"
    # needs some additional points in some case (XXX explain cases)
    __select__ = is_instance("Any") & yes()

    _options = FormRenderer._options + ("main_form_title",)
    main_form_title = _("main informations")

    def open_form(self, form, values):
        attrs_fs_label = ""
        if self.main_form_title:
            attrs_fs_label += (
                '<div class="iformTitle"><span>%s</span></div>'
                % self._cw._(self.main_form_title)
            )
        attrs_fs_label += '<div class="formBody">'
        return attrs_fs_label + super().open_form(form, values)

    def close_form(self, form, values):
        """seems dumb but important for consistency w/ close form, and necessary
        for form renderers overriding open_form to use something else or more than
        and <form>
        """
        return super().close_form(form, values) + "</div>"

    def render_buttons(self, w, form):
        if len(form.form_buttons) == 3:
            w(
                """<table width="100%%">
  <tbody>
   <tr><td align="center">
     %s
   </td><td style="align: right; width: 50%%;">
     %s
     %s
   </td></tr>
  </tbody>
 </table>"""
                % tuple(button.render(form) for button in form.form_buttons)
            )
        else:
            super().render_buttons(w, form)


class EntityInlinedFormRenderer(EntityFormRenderer):
    """This is a specific renderer for entity's form inlined into another
    entity's form.
    """

    __regid__ = "inline"
    fieldset_css_class = "subentity"

    def render_title(self, w, form, values):
        w('<div class="iformTitle">')
        w(
            "<span>%(title)s</span> "
            '#<span class="icounter">%(counter)s</span> ' % values
        )
        if values["removejs"]:
            values["removemsg"] = self._cw._("remove-inlined-entity-form")
            w(
                '[<a href="javascript: %(removejs)s;$.noop();">%(removemsg)s</a>]'
                % values
            )
        w("</div>")

    def render(self, w, form, values):
        form.add_media()
        self.open_form(w, form, values)
        self.render_title(w, form, values)
        # XXX that stinks
        # cleanup values
        for key in ("title", "removejs", "removemsg"):
            values.pop(key, None)
        self.render_fields(w, form, values)
        self.close_form(w, form, values)

    def open_form(self, w, form, values):
        try:
            w('<div id="div-%(divid)s" onclick="%(divonclick)s">' % values)
        except KeyError:
            w('<div id="div-%(divid)s">' % values)
        else:
            w(
                '<div id="notice-%s" class="notice">%s</div>'
                % (
                    values["divid"],
                    self._cw._("click on the box to cancel the deletion"),
                )
            )
        w('<div class="iformBody">')

    def close_form(self, w, form, values):
        w("</div></div>")

    def render_fields(self, w, form, values):
        w('<fieldset id="fs-%(divid)s">' % values)
        fields = self._render_hidden_fields(w, form)
        w("</fieldset>")
        w('<fieldset class="%s">' % self.fieldset_css_class)
        if fields:
            self._render_fields(fields, w, form)
        self.render_child_forms(w, form, values)
        w("</fieldset>")



import sys
import webtest

from cubicweb.pyramid.test import PyramidCWTest, ACCEPTED_ORIGINS


class TestApp(webtest.TestApp):
    def __init__(self, *args, admin_login, admin_password, **kwargs):
        super().__init__(*args, **kwargs)
        self.admin_login = admin_login
        self.admin_password = admin_password
        self._csrf_token = None

    def reset(self):
        super().reset()
        self._csrf_token = None

    def login(self, user=None, password=None, status=303, additional_arguments=None):
        """Log the current http session for the provided credential

        If no user is provided, admin connection are used.
        """
        if user is None:
            user = self.admin_login
            password = self.admin_password
        if password is None:
            password = user

        arguments = {"__login": user, "__password": password}
        if additional_arguments:
            arguments.update(additional_arguments)

        response = self.post("/login", arguments, status=status)

        assert response.status_int == status

        return response

    def logout(self):
        response = self.get("/logout")
        self.reset()
        return response

    def post(
        self,
        route,
        params=None,
        do_not_grab_the_crsf_token=False,
        do_not_inject_origin=False,
        **kwargs,
    ):
        if params is None:
            params = {}

        if (
            isinstance(params, dict)
            and not do_not_grab_the_crsf_token
            and "csrf_token" not in params
        ):
            csrf_token = self.get_csrf_token()

            # "application/json" doesn't submit token in form params but as header value
            if kwargs.get("headers", {}).get("Content-Type") != "application/json":
                if "csrf_token" not in params:
                    params["csrf_token"] = csrf_token
            else:
                if "headers" in kwargs:
                    kwargs["headers"]["X-CSRF-Token"] = csrf_token
                else:
                    kwargs["headers"] = {"X-CSRF-Token": csrf_token}

        if not do_not_inject_origin:
            if "headers" in kwargs and "Origin" not in kwargs["headers"]:
                kwargs["headers"]["Origin"] = "https://" + ACCEPTED_ORIGINS[0]
            elif "headers" not in kwargs:
                kwargs["headers"] = {"Origin": "https://" + ACCEPTED_ORIGINS[0]}

        return super().post(route, params, **kwargs)

    def _inject_params_to_kwargs(
        self, do_not_grab_the_crsf_token, do_not_inject_origin, **kwargs
    ):
        if not do_not_grab_the_crsf_token:
            csrf_token = self.get_csrf_token()
            if "headers" in kwargs:
                kwargs["headers"]["X-CSRF-Token"] = csrf_token
            else:
                kwargs["headers"] = {"X-CSRF-Token": csrf_token}
        if not do_not_inject_origin:
            if "headers" in kwargs and "Origin" not in kwargs["headers"]:
                kwargs["headers"]["Origin"] = "https://" + ACCEPTED_ORIGINS[0]
            elif "headers" not in kwargs:
                kwargs["headers"] = {"Origin": "https://" + ACCEPTED_ORIGINS[0]}
        return kwargs

    def post_json(
        self,
        route,
        params=None,
        do_not_grab_the_crsf_token=False,
        do_not_inject_origin=False,
        **kwargs,
    ):
        kwargs = self._inject_params_to_kwargs(
            do_not_grab_the_crsf_token, do_not_inject_origin, **kwargs
        )
        return super().post_json(route, params, **kwargs)

    def put_json(
        self,
        route,
        params=None,
        do_not_grab_the_crsf_token=False,
        do_not_inject_origin=False,
        **kwargs,
    ):
        kwargs = self._inject_params_to_kwargs(
            do_not_grab_the_crsf_token, do_not_inject_origin, **kwargs
        )
        return super().put_json(route, params, **kwargs)

    def get_csrf_token(self):
        if "csrf_token" in self.cookies:
            return self.cookies["csrf_token"]

        try:
            self.get("/")
        except webtest.app.AppError as e:
            # url doesn't exist or forbidden
            if "404" in e.args[0] or "403" in e.args[0]:
                try:
                    response = self.get("/login")
                except Exception as e:
                    sys.stderr.write(
                        f"ERROR: failed to do a GET on '/' and on '/login' to get the csrf token because: {e}\n"
                    )
                    raise
                else:
                    if "csrf_token" in self.cookies:
                        return self.cookies["csrf_token"]
                    return response.form.fields["csrf_token"][0].value
            else:
                sys.stderr.write(
                    f"ERROR: failed to do a GET on '/' to get the csrf token because: {e}\n"
                )
                raise
        else:
            return self.cookies["csrf_token"]


class PyramidCWTestWithCSRF(PyramidCWTest):
    def includeme(self, config):
        config.registry.settings["pyramid.csrf_trusted_origins"] = ACCEPTED_ORIGINS

    def setUp(self):
        # Skip CubicWebTestTC setUp
        super().setUp()
        self.includeme(self.pyramid_config)
        self.webapp = TestApp(
            self.pyramid_config.make_wsgi_app(),
            extra_environ={"wsgi.url_scheme": "https"},
            admin_login=self.admlogin,
            admin_password=self.admpassword,
        )

    def tearDown(self):
        del self.webapp
        super().tearDown()



# copyright 2010-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Specific views for data sources and related entities (eg CWSource,
CWSourceHostConfig).
"""

import logging

from logilab.common.decorators import cachedproperty

from cubicweb import Unauthorized
from cubicweb import _
from cubicweb.predicates import (
    is_instance,
    score_entity,
    has_related_entities,
    match_user_groups,
    match_kwargs,
    match_view,
    one_line_rset,
)
from cubicweb.utils import make_uid
from cubicweb_web import tags
from cubicweb_web.view import EntityView, StartupView
from cubicweb_web import Redirect, formwidgets as wdgs, facet, action
from cubicweb_web.views import add_etype_button
from cubicweb_web.views import (
    uicfg,
    tabs,
    actions,
    ibreadcrumbs,
    navigation,
    tableview,
    pyviews,
)

_abaa = uicfg.actionbox_appearsin_addmenu
# there are explicit 'add' buttons for those
_abaa.tag_object_of(("CWDataImport", "cw_import_of", "*"), False)

_afs = uicfg.autoform_section
_afs.tag_attribute(("CWSource", "latest_retrieval"), "main", "hidden")
_afs.tag_attribute(("CWSource", "in_synchronization"), "main", "hidden")

_affk = uicfg.autoform_field_kwargs
_affk.tag_attribute(("CWSource", "parser"), {"widget": wdgs.TextInput})

# source primary views #########################################################

_pvs = uicfg.primaryview_section
_pvs.tag_attribute(("CWSource", "name"), "hidden")
_pvs.tag_object_of(("*", "cw_host_config_of", "CWSource"), "hidden")

_pvdc = uicfg.primaryview_display_ctrl
_pvdc.tag_attribute(("CWSource", "type"), {"vid": "attribute"})  # disable reledit

_rc = uicfg.reledit_ctrl
_rc.tag_attribute(("CWSource", "config"), {"rvid": "verbatimattr"})
_rc.tag_attribute(("CWSourceHostConfig", "config"), {"rvid": "verbatimattr"})


class CWSourcePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("CWSource")
    tabs = [_("cwsource-main"), _("cwsource-imports")]
    default_tab = "cwsource-main"


class CWSourceMainTab(tabs.PrimaryTab):
    __regid__ = "cwsource-main"
    __select__ = is_instance("CWSource")

    def render_entity_attributes(self, entity):
        super().render_entity_attributes(entity)
        self.w("cw_host_config_of:%s:subject", entity.eid)
        try:
            hostconfig = self._cw.execute(
                "Any X, XC, XH WHERE X cw_host_config_of S, S eid %(s)s, "
                "X config XC, X match_host XH",
                {"s": entity.eid},
            )
        except Unauthorized:
            pass
        else:
            if hostconfig:
                self.w("<h3>%s</h3>", self._cw._("CWSourceHostConfig_plural"))
                self._cw.view(
                    "table",
                    hostconfig,
                    w=self.w,
                    displaycols=list(range(2)),
                    cellvids={1: "editable-final"},
                )


class CWSourceImportsTab(EntityView):
    __regid__ = "cwsource-imports"
    __select__ = is_instance("CWSource") & has_related_entities(
        "cw_import_of", "object"
    )

    def entity_call(self, entity):
        rset = self._cw.execute(
            "Any X, XST, XET, XS ORDERBY XST DESC WHERE "
            "X cw_import_of S, S eid %(s)s, X status XS, "
            "X start_timestamp XST, X end_timestamp XET",
            {"s": entity.eid},
        )
        self._cw.view("cw.imports-table", rset, w=self.w)


class CWImportsTable(tableview.EntityTableView):
    __regid__ = "cw.imports-table"
    __select__ = is_instance("CWDataImport")
    columns = ["import", "start_timestamp", "end_timestamp"]
    column_renderers = {"import": tableview.MainEntityColRenderer()}
    layout_args = {"display_filter": "top"}


class CWSourceSyncAction(action.Action):
    __regid__ = "cw.source-sync"
    __select__ = (
        action.Action.__select__
        & match_user_groups("managers")
        & one_line_rset()
        & is_instance("CWSource")
        & score_entity(lambda x: x.name != "system")
    )

    title = _("synchronize")
    category = "mainactions"
    order = 20

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid=self.__regid__)


class CWSourceSyncView(EntityView):
    __regid__ = "cw.source-sync"
    __select__ = (
        match_user_groups("managers")
        & one_line_rset()
        & is_instance("CWSource")
        & score_entity(lambda x: x.name != "system")
    )

    title = _("synchronize")

    def entity_call(self, entity):
        import_log_eid = self._cw.call_service("source-sync", source_eid=entity.eid)
        msg = self._cw._(
            "Synchronization has been requested, refresh this page in a few " "minutes."
        )
        import_log = self._cw.entity_from_eid(import_log_eid)
        url = import_log.absolute_url(__message=msg)
        raise Redirect(url)


# sources management view ######################################################


class ManageSourcesAction(actions.ManagersAction):
    __regid__ = "cwsource"
    title = _("data sources")
    category = "manage"
    order = 100


class CWSourcesManagementView(StartupView):
    __regid__ = "cw.sources-management"
    rql = (
        "Any S,ST,SP,SD,SN ORDERBY SN WHERE S is CWSource, S name SN, S type ST, "
        "S latest_retrieval SD, S parser SP"
    )
    title = _("data sources management")

    def call(self, **kwargs):
        self.w("<h1>%s</h1>", self._cw._(self.title))
        self.w(add_etype_button(self._cw, "CWSource"))
        self.w('<div class="clear"></div>')
        self.wview("cw.sources-table", self._cw.execute(self.rql))


class CWSourcesTable(tableview.EntityTableView):
    __regid__ = "cw.sources-table"
    __select__ = is_instance("CWSource")
    columns = ["source", "type", "parser", "latest_retrieval", "latest_import"]

    class LatestImportColRenderer(tableview.EntityTableColRenderer):
        def render_cell(self, w, rownum):
            entity = self.entity(rownum)
            rset = self._cw.execute(
                "Any X,XS,XST ORDERBY XST DESC LIMIT 1 WHERE "
                "X cw_import_of S, S eid %(s)s, X status XS, "
                "X start_timestamp XST",
                {"s": entity.eid},
            )
            if rset:
                self._cw.view("incontext", rset, row=0, w=w)
            else:
                w(self.empty_cell_content)

    column_renderers = {
        "source": tableview.MainEntityColRenderer(),
        "latest_import": LatestImportColRenderer(
            header=_("latest import"), sortable=False
        ),
    }


# datafeed source import #######################################################

REVERSE_SEVERITIES = {
    logging.DEBUG: _("DEBUG"),
    logging.INFO: _("INFO"),
    logging.WARNING: _("WARNING"),
    logging.ERROR: _("ERROR"),
    logging.FATAL: _("FATAL"),
}


def log_to_table(req, rawdata):
    data = []
    for msg_idx, msg in enumerate(rawdata.split("<br/>")):
        record = msg.strip()
        if not record:
            continue
        try:
            severity, url, line, msg = record.split("\t", 3)
        except ValueError:
            req.warning("badly formated log %s" % record)
            url = line = ""
            severity = logging.DEBUG
            msg = record
        data.append((severity, url, line, msg))
    return data


class LogTableLayout(tableview.TableLayout):
    __select__ = match_view("cw.log.table")
    needs_js = tableview.TableLayout.needs_js + ("cubicweb.log.js",)
    needs_css = tableview.TableLayout.needs_css + ("cubicweb.log.css",)
    columns_css = {
        0: "logSeverity",
        1: "logPath",
        2: "logLine",
        3: "logMsg",
    }

    def render_table(self, w, actions, paginate):
        default_level = self.view.cw_extra_kwargs["default_level"]
        if default_level != "Debug":
            self._cw.add_onload(
                '$("select.log_filter").val("%s").change();'
                % self._cw.form.get("logLevel", default_level)
            )
        w('\n<form action="#"><fieldset>')
        w("<label>%s</label>" % self._cw._("Message threshold"))
        w(
            '<select class="log_filter" onchange="filterLog(\'%s\', this.options[this.selectedIndex].value)">'
            % self.view.domid
        )
        for level in ("DEBUG", "INFO", "WARNING", "ERROR", "FATAL"):
            w(
                '<option value="%s">%s</option>'
                % (level.capitalize(), self._cw._(level))
            )
        w("</select>")
        w("</fieldset></form>")
        super().render_table(w, actions, paginate)

    def table_attributes(self):
        attrs = super().table_attributes()
        attrs["id"] = "table" + self.view.domid
        return attrs

    def row_attributes(self, rownum):
        attrs = super().row_attributes(rownum)
        attrs["id"] = "log_msg_%i" % rownum
        severityname = REVERSE_SEVERITIES[int(self.view.pyvalue[rownum][0])]
        attrs["class"] = "log%s" % severityname.capitalize()
        return attrs

    def cell_attributes(self, rownum, colnum, colid):
        attrs = super().cell_attributes(rownum, colnum, colid)
        attrs["class"] = self.columns_css[colnum]
        return attrs


class LogTable(pyviews.PyValTableView):
    __regid__ = "cw.log.table"
    headers = [_("severity"), _("url"), _("line"), _("message")]

    @cachedproperty
    def domid(self):
        return make_uid("logTable")

    class SeverityRenderer(pyviews.PyValTableColRenderer):
        def render_cell(self, w, rownum):
            severity = self.data[rownum][0]
            w(
                '<a class="internallink" href="javascript:;" title="%(title)s" '
                """onclick="document.location.hash='%(msg_id)s';">&#182;</a>"""
                "&#160;%(severity)s"
                % {
                    "severity": self._cw._(REVERSE_SEVERITIES[int(severity)]),
                    "title": self._cw._("permalink to this message"),
                    "msg_id": "log_msg_%i" % rownum,
                }
            )

        def sortvalue(self, rownum):
            return int(self.data[rownum][0])

    class URLRenderer(pyviews.PyValTableColRenderer):
        def render_cell(self, w, rownum):
            url = self.data[rownum][1]
            if url and url.startswith("http"):
                url = tags.a(url, href=url)
            w(url or "&#160;")

    class LineRenderer(pyviews.PyValTableColRenderer):
        def render_cell(self, w, rownum):
            line = self.data[rownum][2]
            w(line or "&#160;")

    class MessageRenderer(pyviews.PyValTableColRenderer):
        snip_over = 7

        def render_cell(self, w, rownum):
            msg = self.data[rownum][3]
            lines = msg.splitlines()
            if len(lines) <= self.snip_over:
                w('<pre class="rawtext">%s</pre>' % msg)
            else:
                # The make_uid argument has no specific meaning here.
                div_snip_id = make_uid("log_snip_")
                div_full_id = make_uid("log_full_")
                divs_id = (div_snip_id, div_full_id)
                snip = "\n".join(
                    (
                        lines[0],
                        lines[1],
                        "  ...",
                        "    %i more lines [double click to expand]" % (len(lines) - 4),
                        "  ...",
                        lines[-2],
                        lines[-1],
                    )
                )
                divs = (
                    (div_snip_id, snip, "expand", "class='collapsed'"),
                    (div_full_id, msg, "collapse", "class='hidden'"),
                )
                for div_id, content, button, h_class in divs:
                    text = self._cw._(button)
                    js = "toggleVisibility('%s'); toggleVisibility('%s');" % divs_id
                    w(f'<div id="{div_id}" {h_class}>')
                    w(
                        '<pre class="raw_test" ondblclick="javascript: %s" '
                        'title="%s" style="display: block;">' % (js, text)
                    )
                    w(content)
                    w("</pre>")
                    w("</div>")

    column_renderers = {
        0: SeverityRenderer(),
        1: URLRenderer(sortable=False),
        2: LineRenderer(sortable=False),
        3: MessageRenderer(sortable=False),
    }


class DataFeedSourceDataImport(EntityView):
    __select__ = EntityView.__select__ & match_kwargs("rtype")
    __regid__ = "cw.formated_log"

    def cell_call(self, row, col, rtype, loglevel="Info", **kwargs):
        if "dispctrl" in self.cw_extra_kwargs:
            loglevel = self.cw_extra_kwargs["dispctrl"].get("loglevel", loglevel)
        entity = self.cw_rset.get_entity(row, col)
        value = getattr(entity, rtype)
        if value:
            self._cw.view(
                "cw.log.table",
                pyvalue=log_to_table(self._cw, value),
                default_level=loglevel,
                w=self.w,
            )
        else:
            self.w(self._cw._("no log to display"))


_pvs.tag_attribute(("CWDataImport", "log"), "relations")
_pvdc.tag_attribute(("CWDataImport", "log"), {"vid": "cw.formated_log"})
_pvs.tag_subject_of(("CWDataImport", "cw_import_of", "*"), "hidden")  # in breadcrumbs
_pvs.tag_object_of(("*", "cw_import_of", "CWSource"), "hidden")  # in dedicated tab


class CWDataImportIPrevNextAdapter(navigation.IPrevNextAdapter):
    __select__ = is_instance("CWDataImport")

    def next_entity(self):
        if self.entity.start_timestamp is not None:
            # add NOT X eid %(e)s because > may not be enough
            rset = self._cw.execute(
                "Any X,XSTS ORDERBY 2 LIMIT 1 WHERE X is CWDataImport, "
                "X cw_import_of S, S eid %(s)s, NOT X eid %(e)s, "
                "X start_timestamp XSTS, X start_timestamp > %(sts)s",
                {
                    "sts": self.entity.start_timestamp,
                    "e": self.entity.eid,
                    "s": self.entity.cwsource.eid,
                },
            )
            if rset:
                return rset.get_entity(0, 0)

    def previous_entity(self):
        if self.entity.start_timestamp is not None:
            # add NOT X eid %(e)s because < may not be enough
            rset = self._cw.execute(
                "Any X,XSTS ORDERBY 2 DESC LIMIT 1 WHERE X is CWDataImport, "
                "X cw_import_of S, S eid %(s)s, NOT X eid %(e)s, "
                "X start_timestamp XSTS, X start_timestamp < %(sts)s",
                {
                    "sts": self.entity.start_timestamp,
                    "e": self.entity.eid,
                    "s": self.entity.cwsource.eid,
                },
            )
            if rset:
                return rset.get_entity(0, 0)


class CWDataImportStatusFacet(facet.AttributeFacet):
    __regid__ = "datafeed.dataimport.status"
    __select__ = is_instance("CWDataImport")
    rtype = "status"


# breadcrumbs configuration ####################################################


class CWsourceConfigIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("CWSourceHostConfig")

    def parent_entity(self):
        return self.entity.cwsource


class CWDataImportIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("CWDataImport")

    def parent_entity(self):
        return self.entity.cw_import_of[0]



# copyright 2017-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""
Url redirection using pyramid
-----------------------------

This module allow to define redirection rules used by pyramid before route selection.

Each rule has to be added with `add_rewriting_rule` method in a pyramid `includeme`
function to be used.

Example of usage :

.. code-block:: ini

    def includeme(config):
        config.add_redirection_rule(r'<a regex>', callback_method, keep_query_components)


The callback_method takes four arguments:

- a pyramid request ;
- the url that has to be matched by the rule ;
- a dictionnary containing all the named subgroups of the regex match ;
- a boolean to decide if we want to keep parameters of the query after the
  redirection (True by default).

It must return a string corresponding to the new url.
"""
import re

from pyramid.httpexceptions import HTTPSeeOther


def add_redirection_rule(config, rule, callback_method, keep_query_component=True):
    """
    Declare a pyramid directive allowing to add a redirection rule.
    Each rewriting rule can then be added in the `includeme` at the end of the file
    by using `config.add_redirection_rule(rule, callback_method`.
    :param config: pyramid configuration, used by pyramid.
    :param rule: a regex used to match uri.
    :param callback_method: a method which return the redirected uri.
    :param keep_query_component: if we want to keep the query component of the URL (the
    part after the first "?") after redirection.
    """
    if not hasattr(config.registry, "redirection_rules"):
        config.registry.redirection_rules = []
    config.registry.redirection_rules.append(
        (re.compile(rule), callback_method, keep_query_component)
    )


def url_redirection_tween_factory(handler, registry):
    """A pyramid tween handler that browse each `redirection_rules` added with
    `add_redirection_rule` directive to find if the current path matches a rule.
    """

    if not hasattr(registry, "redirection_rules"):
        return handler

    def url_redirection_tween(request):
        for (
            rule,
            callback_method,
            keep_query_component,
        ) in request.registry.redirection_rules:
            path = request.path_info
            match = rule.match(path)
            if match:
                redirection_path = callback_method(request, path, match.groupdict())
                if redirection_path and keep_query_component and request.query_string:
                    raise HTTPSeeOther(f"{redirection_path}?{request.query_string}")
                if redirection_path:
                    raise HTTPSeeOther(redirection_path)
        return handler(request)

    return url_redirection_tween


def includeme(config):
    """Add `add_redirection_rule` pyramid directive and `url_redirection_tween`
    pyramid tween handler to the pyramid configuration."""
    config.add_directive("add_redirection_rule", add_redirection_rule)
    config.add_tween(
        "cubicweb_web.pyramid.url_redirection.url_redirection_tween_factory"
    )



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Specific views for users and groups"""


from hashlib import sha1  # pylint: disable=E0611

from cubicweb import _
from cubicweb.predicates import one_line_rset, is_instance, match_user_groups
from cubicweb_web.view import EntityView, StartupView
from cubicweb_web import action, formwidgets
from cubicweb_web import tags
from cubicweb_web.views import uicfg, tabs, tableview, actions, add_etype_button

_pvs = uicfg.primaryview_section
_pvs.tag_attribute(("CWUser", "login"), "hidden")

_affk = uicfg.autoform_field_kwargs
_affk.tag_subject_of(
    ("CWUser", "in_group", "CWGroup"), {"widget": formwidgets.InOutWidget}
)


class UserPreferencesEntityAction(action.Action):
    __regid__ = "prefs"
    __select__ = (
        one_line_rset()
        & is_instance("CWUser")
        & match_user_groups("owners", "managers")
    )

    title = _("preferences")
    category = "mainactions"

    def url(self):
        user = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return user.absolute_url(vid="propertiesform")


class FoafView(EntityView):
    __regid__ = "foaf"
    __select__ = is_instance("CWUser")

    title = _("foaf")
    templatable = False
    content_type = "application/rdf+xml"

    def call(self):
        self.w(
            """<?xml version="1.0" encoding="%s"?>
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
         xmlns:foaf="http://xmlns.com/foaf/0.1/"> """,
            self._cw.encoding,
        )
        for i in range(self.cw_rset.rowcount):
            self.cell_call(i, 0)
        self.w("</rdf:RDF>\n")

    def entity_call(self, entity, **kwargs):
        entity.complete()
        # account
        self.w('<foaf:OnlineAccount rdf:about="%s">\n', entity.absolute_url())
        self.w("  <foaf:accountName>%s</foaf:accountName>\n", entity.login)
        self.w("</foaf:OnlineAccount>\n")
        # person
        self.w('<foaf:Person rdf:about="%s#user">\n', entity.absolute_url())
        self.w('  <foaf:account rdf:resource="%s" />\n', entity.absolute_url())
        if entity.surname:
            self.w("<foaf:familyName>%s</foaf:familyName>\n", entity.surname)
        if entity.firstname:
            self.w("<foaf:givenName>%s</foaf:givenName>\n", entity.firstname)
        emailaddr = entity.cw_adapt_to("IEmailable").get_email()
        if emailaddr:
            self.w(
                "<foaf:mbox_sha1sum>%s</foaf:mbox_sha1sum>\n",
                sha1(emailaddr.encode("utf-8")).hexdigest(),
            )
        self.w("</foaf:Person>\n")


# group views ##################################################################

_pvs.tag_attribute(("CWGroup", "name"), "hidden")
_pvs.tag_subject_of(("CWGroup", "read_permission", "*"), "relations")
_pvs.tag_subject_of(("CWGroup", "add_permission", "*"), "relations")
_pvs.tag_subject_of(("CWGroup", "delete_permission", "*"), "relations")
_pvs.tag_subject_of(("CWGroup", "update_permission", "*"), "relations")
_pvs.tag_object_of(("CWUser", "in_group", "CWGroup"), "hidden")
_pvs.tag_object_of(("*", "require_group", "CWGroup"), "hidden")


class CWGroupPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("CWGroup")
    tabs = [_("cwgroup-main"), _("cwgroup-permissions")]
    default_tab = "cwgroup-main"


class CWGroupMainTab(tabs.PrimaryTab):
    __regid__ = "cwgroup-main"
    __select__ = tabs.PrimaryTab.__select__ & is_instance("CWGroup")

    def render_entity_attributes(self, entity):
        rset = self._cw.execute(
            "Any U, FN, LN, CD, LL ORDERBY L WHERE U in_group G, "
            "U login L, U firstname FN, U surname LN, U creation_date CD, "
            "U last_login_time LL, G eid %(x)s",
            {"x": entity.eid},
        )
        self.wview("cwgroup.users", rset, "null")


class CWGroupUsersTable(tableview.RsetTableView):
    __regid__ = "cwgroup.users"
    __select__ = is_instance("CWUser")
    headers = (
        _("user"),
        _("first name"),
        _("last name"),
        _("creation date"),
        _("last login time"),
    )
    layout_args = {"display_filter": "top"}
    finalvid = "editable-final"


class CWGroupPermTab(EntityView):
    __regid__ = "cwgroup-permissions"
    __select__ = is_instance("CWGroup")

    def entity_call(self, entity):
        self._cw.add_css(("cubicweb.schema.css", "cubicweb.acl.css"))
        objtype_access = {
            "CWEType": ("read", "delete", "add", "update"),
            "CWRelation": ("add", "delete"),
        }
        rql_cwetype = (
            "DISTINCT Any X WHERE X %s_permission CWG, X is CWEType, " "CWG eid %%(e)s"
        )
        rql_cwrelation = (
            "DISTINCT Any RT WHERE X %s_permission CWG, X is CWRelation, "
            "X relation_type RT, CWG eid %%(e)s"
        )
        self.render_objtype_access(entity, "CWEType", objtype_access, rql_cwetype)
        self.render_objtype_access(entity, "CWRelation", objtype_access, rql_cwrelation)

    def render_objtype_access(self, entity, objtype, objtype_access, rql):
        self.w("<h4>%s</h4>", self._cw._(objtype))
        for access_type in objtype_access[objtype]:
            rset = self._cw.execute(rql % access_type, {"e": entity.eid})
            if rset:
                self.w(
                    "<div>%s:</div>",
                    self._cw.__(access_type + "_permission"),
                    escape=False,
                )
                self.w(
                    "<div>%s</div><br/>",
                    self._cw.view("csv", rset, "null"),
                    escape=False,
                )


class CWGroupInContextView(EntityView):
    __regid__ = "incontext"
    __select__ = is_instance("CWGroup")

    def entity_call(self, entity, **kwargs):
        entity.complete()
        self.w(
            '<a href="%s" class="%s">%s</a>',
            entity.absolute_url(),
            entity.name,
            entity.printable_value("name"),
        )


# user / groups management views ###############################################


class ManageUsersAction(actions.ManagersAction):
    __regid__ = "cwuser"  # see rewrite rule /cwuser
    title = _("users and groups")
    category = "manage"


class UsersAndGroupsManagementView(tabs.TabsMixin, StartupView):
    __regid__ = "cw.users-and-groups-management"
    __select__ = StartupView.__select__ & match_user_groups("managers")
    title = _("Users and groups management")
    tabs = [_("cw.users-management"), _("cw.groups-management")]
    default_tab = "cw.users-management"

    def call(self, **kwargs):
        """The default view representing the instance's management"""
        self.w("<h1>%s</h1>", self._cw._(self.title))
        self.render_tabs(self.tabs, self.default_tab)


class CWUserManagementView(StartupView):
    __regid__ = "cw.users-management"
    __select__ = StartupView.__select__ & match_user_groups("managers")
    cache_max_age = 0  # disable caching
    # XXX one could wish to display for instance only user's firstname/surname
    # for non managers but filtering out NULL caused crash with an ldapuser
    # source. The ldapuser source has been dropped and this code can be updated.
    rql = (
        "Any U,US,F,S,U,UAA,UDS, L,UAA,USN,UDSN ORDERBY L WHERE U is CWUser, "
        "U login L, U firstname F, U surname S, "
        "U in_state US, US name USN, "
        "U primary_email UA?, UA address UAA, "
        "U cw_source UDS, US name UDSN"
    )

    def call(self, **kwargs):
        self.w(add_etype_button(self._cw, "CWUser"))
        self.w('<div class="clear"></div>')
        self.wview("cw.users-table", self._cw.execute(self.rql))


class CWUsersTable(tableview.EntityTableView):
    __regid__ = "cw.users-table"
    __select__ = is_instance("CWUser")
    columns = [
        "user",
        "in_state",
        "firstname",
        "surname",
        "in_group",
        "primary_email",
        "cw_source",
    ]
    layout_args = {"display_filter": "top"}
    finalvid = "editable-final"

    column_renderers = {
        "user": tableview.EntityTableColRenderer(
            renderfunc=lambda w, x: w(tags.a(x.login, href=x.absolute_url())),
            sortfunc=lambda x: x.login,
        ),
        "in_state": tableview.EntityTableColRenderer(
            renderfunc=lambda w, x: w(x.cw_adapt_to("IWorkflowable").printable_state),
            sortfunc=lambda x: x.cw_adapt_to("IWorkflowable").printable_state,
        ),
        "in_group": tableview.EntityTableColRenderer(
            renderfunc=lambda w, x: x.view(
                "reledit", rtype="in_group", role="subject", w=w
            )
        ),
        "primary_email": tableview.RelatedEntityColRenderer(
            getrelated=lambda x: x.primary_email and x.primary_email[0] or None
        ),
        "cw_source": tableview.RelatedEntityColRenderer(
            getrelated=lambda x: x.cw_source[0]
        ),
    }


class CWGroupsManagementView(StartupView):
    __regid__ = "cw.groups-management"
    __select__ = StartupView.__select__ & match_user_groups("managers")
    cache_max_age = 0  # disable caching
    rql = 'Any G,GN ORDERBY GN WHERE G is CWGroup, G name GN, NOT G name "owners"'

    def call(self, **kwargs):
        self.w(add_etype_button(self._cw, "CWGroup"))
        self.w('<div class="clear"></div>')
        self.wview("cw.groups-table", self._cw.execute(self.rql))


class CWGroupsTable(tableview.EntityTableView):
    __regid__ = "cw.groups-table"
    __select__ = is_instance("CWGroup")
    columns = ["group", "nb_users"]
    layout_args = {"display_filter": "top"}

    column_renderers = {
        "group": tableview.MainEntityColRenderer(),
        "nb_users": tableview.EntityTableColRenderer(
            header=_("num. users"),
            renderfunc=lambda w, x: w(str(x.num_users())),
            sortfunc=lambda x: x.num_users(),
        ),
    }



"""define default ui properties"""

# CSS stylesheets to include systematically in HTML headers
# use the following line if you *need* to keep the old stylesheet
STYLESHEETS = [
    data("cubicweb.css"),
]
STYLESHEETS_PRINT = [data("cubicweb.print.css")]

# Javascripts files to include systematically in HTML headers
JAVASCRIPTS = [
    data("jquery.js"),
    data("jquery-migrate.js"),
    data("cubicweb.js"),
    data("cubicweb.compat.js"),
    data("cubicweb.python.js"),
    data("cubicweb.htmlhelpers.js"),
]

# where is installed fckeditor
FCKEDITOR_PATH = "/usr/share/fckeditor/"

# favicon and logo for the instance
FAVICON = data("favicon.ico")
LOGO = data("logo.png")

# rss logo (link to get the rss view of a selection)
RSS_LOGO = data("rss.png")
RSS_LOGO_16 = data("feed-icon16x16.png")
RSS_LOGO_32 = data("feed-icon32x32.png")

# XXX cleanup resources below, some of them are probably not used
# (at least entity types icons...)

# images
HELP = data("help.png")
SEARCH_GO = data("go.png")
PUCE_UP = data("puce_up.png")
PUCE_DOWN = data("puce_down.png")

# button icons
OK_ICON = data("ok.png")
CANCEL_ICON = data("cancel.png")
APPLY_ICON = data("plus.png")
TRASH_ICON = data("trash_can_small.png")

# icons for entity types
BOOKMARK_ICON = data("icon_bookmark.gif")
EMAILADDRESS_ICON = data("icon_emailaddress.gif")
EUSER_ICON = data("icon_euser.gif")
STATE_ICON = data("icon_state.gif")

# other icons
CALENDAR_ICON = data("calendar.gif")
CANCEL_EMAIL_ICON = data("sendcancel.png")
SEND_EMAIL_ICON = data("sendok.png")
DOWNLOAD_ICON = data("download.gif")
UPLOAD_ICON = data("upload.gif")
GMARKER_ICON = data("gmap_blue_marker.png")
UP_ICON = data("up.gif")

# colors, fonts, etc

# default (body, html)
defaultColor = "#000"
defaultFontFamily = "'Bitstream Vera Sans','Lucida Grande','Lucida Sans Unicode','Geneva','Verdana',sans-serif"
defaultSize = "12px"
defaultLineHeight = "1.5"
defaultLineHeightEm = lazystr("%(defaultLineHeight)sem")

inputHeight = "1.3em"
inputPadding = "O.2em"
# XXX
defaultLayoutMargin = "8px"

# header
headerBgColor = "#ff7700"
headerBg = lazystr('%(headerBgColor)s url("banner.png") repeat-x top left')


# scale 3:5 stranded
# h1 { font-size:2em; }
# h2 { font-size:1.61538em; }
# h3 { font-size:1.23077em; }
#
# scale le corbusier */
# h1 { font-size:2.11538em; }
# h2 { font-size:1.61538em; }
# h3 { font-size:1.30769em; }

# h
h1FontSize = "2.3em"  # 25.3833px
h1Padding = "0 0 0.14em 0 "
h1Margin = "0.8em 0 0.5em"
h1Color = "#000"
h1BorderBottomStyle = lazystr("0.06em solid %(h1Color)s")

h2FontSize = "2em"  #
h2Padding = "0.4em 0 0.35em 0"  # 22.0667px
h2Margin = "0"

h3FontSize = "1.7em"  # 18.75px
h3Padding = "0.5em 0 0.57em 0"
h3Margin = "0"

h4FontSize = "1.4em"  # 15.45px

# links
aColor = "#e6820e"


# page frame
pageBgColor = "#e2e2e2"
pageContentBorderColor = "#ccc"
pageContentBgColor = "#fff"
pageContentPadding = "1em"
pageMinHeight = "800px"

# boxes ########################################################################

# title of contextFree / contextual boxes
contextFreeBoxTitleBgColor = "#CFCEB7"
contextFreeBoxTitleBg = lazystr(
    '%(contextFreeBoxTitleBgColor)s url("contextFreeBoxHeader.png") repeat-x 50%% 50%%'
)
contextFreeBoxTitleColor = "#000"

contextualBoxTitleBgColor = "#FF9900"
contextualBoxTitleBg = lazystr(
    '%(contextualBoxTitleBgColor)s url("contextualBoxHeader.png") repeat-x 50%% 50%%'
)
contextualBoxTitleColor = "#FFF"

# title of 'incontext' boxes (eg displayed inside the primary view)
incontextBoxTitleBgColor = lazystr("%(contextFreeBoxTitleBgColor)s")
incontextBoxTitleBg = lazystr(
    '%(incontextBoxTitleBgColor)s url("incontextBoxHeader.png") repeat-x 50%% 50%%'
)
incontextBoxTitleColor = "#000"

# body of boxes in the left or right column (whatever contextFree / contextual)
leftrightBoxBodyBgColor = "#FFF"
leftrightBoxBodyBg = lazystr("%(leftrightBoxBodyBgColor)s")
leftrightBoxBodyColor = "black"
leftrightBoxBodyHoverBgColor = "#EEEDD9"

# body of 'incontext' boxes (eg displayed insinde the primary view)
incontextBoxBodyBgColor = "#f8f8ee"
incontextBoxBodyBg = lazystr("%(incontextBoxBodyBgColor)s")
incontextBoxBodyColor = "#555544"
incontextBoxBodyHoverBgColor = lazystr("%(incontextBoxBodyBgColor)s")


# table listing & co ###########################################################
listingBorderColor = "#ccc"
listingHeaderBgColor = "#efefef"
listingHighlightedBgColor = "#fbfbfb"

# puce
bulletDownImg = 'url("puce_down.png") 98% 6px no-repeat'

# forms
formHeaderBgColor = lazystr("%(listingHeaderBgColor)s")
helperColor = "#555"

# button
buttonBorderColor = "#edecd2"
buttonBgColor = "#fffff8"
buttonBgImg = 'url("button.png") repeat-x 50% 50%'

# messages
msgBgColor = "#f8f8ee"
infoMsgBgImg = 'url("information.png") 5px center no-repeat'
errorMsgBgImg = 'url("error.png") 100% 50% no-repeat'
errorMsgColor = "#ed0d0d"

# facets
facet_titleFont = "bold SansSerif"
facet_Padding = ".4em"
facet_MarginBottom = ".4em"
facet_vocabMaxHeight = (
    "12em"  # ensure < FACET_GROUP_HEIGHT by some const. factor (e.g 3em)
)



# copyright 2014-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
from cubicweb_web.devtools.testlib import WebCWTC


class CSVExportViewsCWTC(WebCWTC):
    def test_csvexport(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
                "Any GN,COUNT(X) GROUPBY GN ORDERBY GN " "WHERE X in_group G, G name GN"
            )
            data = self.view("csvexport", rset, req=req)
            self.assertEqual(
                req.headers_out.getRawHeaders("content-type"),
                ["text/comma-separated-values;charset=UTF-8"],
            )
            expected_data = "String;COUNT(CWUser)\nguests;1\nmanagers;1"
            self.assertMultiLineEqual(expected_data, data.decode("utf-8"))

    def test_csvexport_on_empty_rset(self):
        """Should return the CSV header."""
        with self.admin_access.web_request() as req:
            rset = req.execute(
                "Any GN,COUNT(X) GROUPBY GN ORDERBY GN "
                'WHERE X in_group G, G name GN, X login "Miles"'
            )
            data = self.view("csvexport", rset, req=req)
            self.assertEqual(
                req.headers_out.getRawHeaders("content-type"),
                ["text/comma-separated-values;charset=UTF-8"],
            )
            expected_data = "String;COUNT(CWUser)"
            self.assertMultiLineEqual(expected_data, data.decode("utf-8"))


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common.testlib import unittest_main

from cubicweb.devtools.htmlparser import XMLValidator

from cubicweb_web.devtools.testlib import WebCWTC


class TreeViewCWTC(WebCWTC):
    def test_treeview(self):
        with self.admin_access.repo_cnx() as cnx:
            ce = cnx.create_entity
            root = ce("TreeNode", name="root")
            node = ce("TreeNode", name="node1", parent=root)
            ce("TreeNode", name="leaf1a", parent=node)
            ce("TreeNode", name="leaf1b", parent=node)
            node = ce("TreeNode", name="node2", parent=root)
            ce("TreeNode", name="leaf2a", parent=node)
            ce("TreeNode", name="leaf2b", parent=node)
            root_eid = root.eid
            cnx.commit()

        with self.admin_access.web_request() as req:
            root = req.entity_from_eid(root_eid)
            valid = self.content_type_validators.get("text/html", XMLValidator)()
            page = valid.parse_string(root.view("tree", klass="oh-my-class"))
            uls = page.find_tag("ul", gettext=False)
            for _, attrib in uls:
                self.assertEqual(attrib["class"], "oh-my-class")


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Specific views for CWProperty (eg site/user preferences"""


from logilab.common.decorators import cached
from logilab.mtconverter import xml_escape

from cubicweb import UnknownProperty
from cubicweb import _
from cubicweb.predicates import (
    one_line_rset,
    none_rset,
    is_instance,
    match_user_groups,
    logged_user_in_rset,
)
from cubicweb_web.view import StartupView
from cubicweb_web.form import FormViewMixIn
from cubicweb_web.formfields import FIELDS, StringField
from cubicweb_web.formwidgets import (
    Select,
    TextInput,
    SubmitButton,
    FieldWidget,
)
from cubicweb_web.views import uicfg, primary, formrenderers, editcontroller
from cubicweb_web.views.ajaxcontroller import ajaxfunc

uicfg.primaryview_section.tag_object_of(("*", "for_user", "*"), "hidden")

# some string we want to be internationalizable for nicer display of property
# groups
_("navigation")
_("ui")
_("boxes")
_("components")
_("ctxcomponents")
_("navigation.combobox-limit")
_("navigation.page-size")
_("navigation.related-limit")
_("navigation.short-line-size")
_("ui.date-format")
_("ui.datetime-format")
_("ui.default-text-format")
_("ui.fckeditor")
_("ui.float-format")
_("ui.language")
_("ui.time-format")
_("open all")
_("ui.main-template")
_("ui.site-title")
_("ui.encoding")
_("category")


def make_togglable_link(nodeid, label):
    """builds a HTML link that switches the visibility & remembers it"""
    return f"<a href=\"javascript: togglePrefVisibility('{nodeid}')\">{label}</a>"


def css_class(someclass):
    return someclass and 'class="%s"' % someclass or ""


class CWPropertyPrimaryView(primary.PrimaryView):
    __select__ = is_instance("CWProperty")
    skip_none = False


class SystemCWPropertiesForm(FormViewMixIn, StartupView):
    """site-wide properties edition form"""

    __regid__ = "systempropertiesform"
    __select__ = none_rset() & match_user_groups("managers")
    form_buttons = [SubmitButton()]

    title = _("site configuration")
    category = "startupview"

    def linkable(self):
        return True

    def url(self):
        """return the url associated with this view. We can omit rql here"""
        return self._cw.build_url("view", vid=self.__regid__)

    def _cookie_name(self, somestr):
        return str(f"{self._cw.vreg.config.appid}_property_{somestr}")

    def _group_status(self, group, default="hidden"):
        """return css class name 'hidden' (collapsed), or '' (open)"""
        cookies = self._cw.get_cookie()
        cookiename = self._cookie_name(group)
        cookie = cookies.get(cookiename)
        if cookie is None:
            self._cw.set_cookie(cookiename, default, maxage=None)
            status = default
        else:
            status = cookie.value
        return status

    def call(self, **kwargs):
        self._cw.add_js(
            ("cubicweb.preferences.js", "cubicweb.edition.js", "cubicweb.ajax.js")
        )
        self._cw.add_css("cubicweb.preferences.css")
        mainopts, groupedopts = self.group_properties()
        # precompute all forms first to consume error message
        mainforms, groupedforms = self.build_forms(mainopts, groupedopts)
        _ = self._cw._
        self.w("<h1>%s</h1>\n", _(self.title))
        for label, group, form in sorted((_(g), g, f) for g, f in mainforms.items()):
            self.wrap_main_form(group, label, form)
        for label, group, objects in sorted(
            (_(g), g, o) for g, o in groupedforms.items()
        ):
            self.wrap_grouped_form(group, label, objects)

    @property
    @cached
    def cwprops_rset(self):
        return self._cw.execute(
            "Any P,K,V WHERE P is CWProperty, P pkey K, P value V, NOT P for_user U"
        )

    @property
    def defined_keys(self):
        values = {}
        for i, entity in enumerate(self.cwprops_rset.entities()):
            values[entity.pkey] = i
        return values

    def group_properties(self):
        mainopts, groupedopts = {}, {}
        vreg = self._cw.vreg
        # "self._regid__=='systempropertiesform'" to skip site wide properties on
        # user's preference but not site's configuration
        for key in vreg.user_property_keys(self.__regid__ == "systempropertiesform"):
            parts = key.split(".")
            if parts[0] in vreg and len(parts) >= 3:
                # appobject configuration
                reg = parts[0]
                oid = ".".join(parts[1:-1])
                groupedopts.setdefault(reg, {}).setdefault(oid, []).append(key)
            else:
                mainopts.setdefault(parts[0], []).append(key)
        return mainopts, groupedopts

    def build_forms(self, mainopts, groupedopts):
        mainforms, groupedforms = {}, {}
        for group, keys in mainopts.items():
            mainforms[group] = self.form(group, keys, False)
        for group, objects in groupedopts.items():
            groupedforms[group] = {}
            for oid, keys in objects.items():
                groupedforms[group][oid] = self.form(group + "_" + oid, keys, True)
        return mainforms, groupedforms

    def entity_for_key(self, key):
        values = self.defined_keys
        if key in values:
            entity = self.cwprops_rset.get_entity(values[key], 0)
        else:
            entity = self._cw.vreg["etypes"].etype_class("CWProperty")(self._cw)
            entity.eid = next(self._cw.varmaker)
            entity.cw_attr_cache["pkey"] = key
            entity.cw_attr_cache["value"] = self._cw.vreg.property_value(key)
        return entity

    def form(self, formid, keys, splitlabel=False):
        form = self._cw.vreg["forms"].select(
            "composite",
            self._cw,
            domid=formid,
            action=self._cw.build_url(),
            form_buttons=self.form_buttons,
            onsubmit="return validatePrefsForm('%s')" % formid,
            submitmsg=self._cw._("changes applied"),
        )
        path = self._cw.relative_path()
        if "?" in path:
            path, params = path.split("?", 1)
            form.add_hidden("__redirectparams", params)
        form.add_hidden("__redirectpath", path)
        for key in keys:
            self.form_row(form, key, splitlabel)
        renderer = self._cw.vreg["formrenderers"].select(
            "cwproperties", self._cw, display_progress_div=False
        )
        data = []
        form.render(w=data.append, renderer=renderer)
        return "\n".join(data)

    def form_row(self, form, key, splitlabel):
        entity = self.entity_for_key(key)
        if splitlabel:
            label = key.split(".")[-1]
        else:
            label = key
        subform = self._cw.vreg["forms"].select(
            "base", self._cw, entity=entity, mainform=False
        )
        subform.append_field(
            PropertyValueField(name="value", label=label, role="subject", eidparam=True)
        )
        subform.add_hidden("pkey", key, eidparam=True, role="subject")
        form.add_subform(subform)
        return subform

    def wrap_main_form(self, group, label, form):
        status = css_class(self._group_status(group))
        self.w(
            '<div class="propertiesform">%s</div>\n',
            make_togglable_link("fieldset_" + group, label),
            escape=False,
        )
        self.w('<div id="fieldset_%s" %s>', group, status, escape=False)
        self.w('<fieldset class="preferences">')
        self.w(form)
        self.w("</fieldset></div>")

    def wrap_grouped_form(self, group, label, objects):
        status = css_class(self._group_status(group))
        self.w(
            '<div class="propertiesform">%s</div>\n',
            make_togglable_link("fieldset_" + group, label),
            escape=False,
        )
        self.w('<div id="fieldset_%s" %s>', group, status, escape=False)
        sorted_objects = sorted(
            (self._cw.__(f"{group}_{o}"), o, f) for o, f in objects.items()
        )
        for label, oid, form in sorted_objects:
            self.wrap_object_form(group, oid, label, form)
        self.w("</div>")

    def wrap_object_form(self, group, oid, label, form):
        self.w('<div class="component">')
        self.w(
            """<div class="componentLink"><a href="#"
                   onclick="javascript:toggleVisibility('field_%s_%s')"
                   class="componentTitle">%s</a>""",
            label,
            oid,
            self._cw._(oid),
        )
        self.w(
            """ (<div class="openlink"><a href="#"
                onclick="javascript:openFieldset('fieldset_%s')">%s</a></div>)""",
            group,
            self._cw._("open all"),
        )
        self.w("</div>")
        docmsgid = f"{group}_{oid}_description"
        doc = self._cw._(docmsgid)
        if doc != docmsgid:
            self.w('<div class="helper">%s</div>', xml_escape(doc).capitalize())
        self.w("</div>")
        self.w(
            '<fieldset id="field_%s_%s" class="%s preferences hidden">',
            label,
            oid,
            group,
        )
        self.w(form)
        self.w("</fieldset>")


class CWPropertiesForm(SystemCWPropertiesForm):
    """user's preferences properties edition form"""

    __regid__ = "propertiesform"
    __select__ = (
        (none_rset() & match_user_groups("users", "managers"))
        | (one_line_rset() & match_user_groups("users") & logged_user_in_rset())
        | (one_line_rset() & match_user_groups("managers") & is_instance("CWUser"))
    )

    title = _("user preferences")

    @property
    def user(self):
        if self.cw_rset is None:
            return self._cw.user
        return self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)

    @property
    @cached
    def cwprops_rset(self):
        return self._cw.execute(
            "Any P,K,V WHERE P is CWProperty, P pkey K, P value V,"
            "P for_user U, U eid %(x)s",
            {"x": self.user.eid},
        )

    def form_row(self, form, key, splitlabel):
        subform = super().form_row(form, key, splitlabel)
        # if user is in the managers group and the property is being created,
        # we have to set for_user explicitly
        if not subform.edited_entity.has_eid() and self.user.matching_groups(
            "managers"
        ):
            subform.add_hidden("for_user", self.user.eid, eidparam=True, role="subject")
        return subform


# cwproperty form objects ######################################################


class PlaceHolderWidget(FieldWidget):
    def render(self, form, field, renderer):
        domid = field.dom_id(form)
        # empty span as well else html validation fail (label is refering to
        # this id)
        return '<div id="div:{}"><span id="{}">{}</span></div>'.format(
            domid,
            domid,
            form._cw._("select a key first"),
        )


class NotEditableWidget(FieldWidget):
    def __init__(self, value, msg=None):
        self.value = value
        self.msg = msg

    def render(self, form, field, renderer):
        domid = field.dom_id(form)
        value = f'<span class="value" id="{domid}">{self.value}</span>'
        if self.msg:
            value += '<div class="helper">%s</div>' % self.msg
        return value


class PropertyKeyField(StringField):
    """specific field for CWProperty.pkey to set the value widget according to
    the selected key
    """

    widget = Select

    def render(self, form, renderer):
        wdg = self.get_widget(form)
        # pylint: disable=E1101
        wdg.attrs["onchange"] = "javascript:setPropValueWidget('%s')" % (
            form.edited_entity.eid
        )
        return wdg.render(form, self, renderer)

    def vocabulary(self, form):
        entity = form.edited_entity
        _ = form._cw._
        if entity.has_eid():
            return [(_(entity.pkey), entity.pkey)]
        choices = entity._cw.vreg.user_property_keys()
        return [("", "")] + sorted(zip((_(v) for v in choices), choices))


class PropertyValueField(StringField):
    """specific field for CWProperty.value  which will be different according to
    the selected key type and vocabulary information
    """

    widget = PlaceHolderWidget

    def render(self, form, renderer=None):
        wdg = self.get_widget(form)
        return wdg.render(form, self, renderer)

    def form_init(self, form):
        entity = form.edited_entity
        if not (entity.has_eid() or "pkey" in entity.cw_attr_cache):
            # no key set yet, just include an empty div which will be filled
            # on key selection
            return
        try:
            pdef = form._cw.vreg.property_info(entity.pkey)
        except UnknownProperty as ex:
            form.warning(
                "%s (you should probably delete that property " "from the database)", ex
            )
            msg = form._cw._("you should probably delete that property")
            self.widget = NotEditableWidget(
                entity.printable_value("value"), f"{msg} ({ex})"
            )
            return
        if entity.pkey.startswith("system."):
            msg = form._cw._("value associated to this key is not editable " "manually")
            self.widget = NotEditableWidget(entity.printable_value("value"), msg)
        # XXX race condition when used from CWPropertyForm, should not rely on
        # instance attributes
        self.value = pdef["default"]
        self.help = pdef["help"]
        vocab = pdef["vocabulary"]
        if vocab is not None:
            if callable(vocab):
                # list() just in case its a generator function
                self.choices = list(vocab())
            else:
                self.choices = vocab
            wdg = Select()
        elif pdef["type"] == "String":  # else we'll get a TextArea by default
            wdg = TextInput()
        else:
            field = FIELDS[pdef["type"]]()
            wdg = field.widget
            if pdef["type"] == "Boolean":
                self.choices = field.vocabulary(form)
        self.widget = wdg


class CWPropertiesFormRenderer(formrenderers.FormRenderer):
    """specific renderer for properties"""

    __regid__ = "cwproperties"

    def open_form(self, form, values):
        err = '<div class="formsg"></div>'
        return super().open_form(form, values) + err

    def _render_fields(self, fields, w, form):
        for field in fields:
            w('<div class="preffield">\n')
            if self.display_label:
                w("%s" % self.render_label(form, field))
            error = form.field_error(field)
            if error:
                w('<span class="error">%s</span>' % error)
            w("%s" % self.render_help(form, field))
            w('<div class="prefinput">')
            w(field.render(form, self))
            w("</div>")
            w("</div>")

    def render_buttons(self, w, form):
        w("<div>\n")
        for button in form.form_buttons:
            w("%s\n" % button.render(form))
        w("</div>")


class CWPropertyIEditControlAdapter(editcontroller.IEditControlAdapter):
    __select__ = is_instance("CWProperty")

    def after_deletion_path(self):
        """return (path, parameters) which should be used as redirect
        information when this entity is being deleted
        """
        return "view", {}


@ajaxfunc(output_type="xhtml")
def prop_widget(self, propkey, varname):
    """specific method for CWProperty handling"""
    entity = self._cw.vreg["etypes"].etype_class("CWProperty")(self._cw)
    entity.eid = varname
    entity.pkey = propkey
    form = self._cw.vreg["forms"].select("edition", self._cw, entity=entity)
    form.build_context()
    vfield = form.field_by_name("value", "subject")
    renderer = formrenderers.FormRenderer(self._cw)
    return vfield.render(form, renderer) + renderer.render_help(form, vfield)


_afs = uicfg.autoform_section
_afs.tag_subject_of(("*", "for_user", "*"), "main", "hidden")
_afs.tag_object_of(("*", "for_user", "*"), "main", "hidden")
_aff = uicfg.autoform_field
_aff.tag_attribute(("CWProperty", "pkey"), PropertyKeyField)
_aff.tag_attribute(("CWProperty", "value"), PropertyValueField)



from cubicweb.devtools import BASE_URL

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views.xmlrss import SERIALIZERS


class EntityXMLViewCWTC(WebCWTC):
    """see also cw.sobjects.test.unittest_parsers"""

    def test(self):
        rels = [
            "tags-object",
            "in_group-subject",
            "in_state-subject",
            "use_email-subject",
        ]
        with self.admin_access.web_request(relation=rels) as req:
            self.assertMultiLineEqual(
                req.user.view("xml"),
                f"""\
<CWUser eid="6" cwuri="{BASE_URL}6" cwsource="system">
  <creation_date>%(cdate)s</creation_date>
  <firstname/>
  <last_login_time/>
  <login>admin</login>
  <modification_date>%(mdate)s</modification_date>
  <surname/>
  <upassword/>
  <tags role="object">
  </tags>
  <in_group role="subject">
    <CWGroup eid="%(group_eid)s" cwuri="{BASE_URL}%(group_eid)s"/>
  </in_group>
  <in_state role="subject">
    <State eid="%(state_eid)s" cwuri="{BASE_URL}%(state_eid)s" name="activated"/>
  </in_state>
  <use_email role="subject">
  </use_email>
</CWUser>
"""
                % {
                    "cdate": SERIALIZERS["Datetime"](req.user.creation_date),
                    "mdate": SERIALIZERS["Datetime"](req.user.modification_date),
                    "state_eid": req.user.in_state[0].eid,
                    "group_eid": req.user.in_group[0].eid,
                },
            )


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb.web.webconfig unit tests"""

import os
from os import path
from unittest import TestCase

from cubicweb_web.devtools.testlib import WebApptestConfiguration


class WebconfigTC(TestCase):
    def setUp(self):
        # need explicit None if dirname(__file__) is empty, see
        # ApptestConfiguration.__init__
        self.config = WebApptestConfiguration("data", __file__)
        self.config._cubes = ["file"]
        self.config.load_configuration()

    def test_nonregr_print_css_as_list(self):
        """make sure PRINT_CSS *must* is a list"""
        config = self.config
        print_css = config.uiprops["STYLESHEETS_PRINT"]
        self.assertIsInstance(print_css, list)

    def test_locate_resource(self):
        self.assertIn("FILE_ICON", self.config.uiprops)
        rname = self.config.uiprops["FILE_ICON"].replace(self.config.datadir_url, "")
        self.assertIn(
            "cubicweb_file", self.config.locate_resource(rname)[0].split(os.sep)
        )
        cubicwebcsspath = self.config.locate_resource("cubicweb.css")[0].split(os.sep)

        # 'shared' if tests under apycot
        self.assertTrue(
            "web" in cubicwebcsspath or "shared" in cubicwebcsspath,
            'neither "web" nor "shared" found in cubicwebcsspath (%s)'
            % cubicwebcsspath,
        )

    def test_locate_all_files(self):
        wdocfiles = list(self.config.locate_all_files("toc.xml"))
        for fpath in wdocfiles:
            self.assertTrue(path.exists(fpath), fpath)
        for expected in [
            path.join("cubicweb_file", "wdoc", "toc.xml"),
            path.join("cubicweb_web", "wdoc", "toc.xml"),
        ]:
            for fpath in wdocfiles:
                if fpath.endswith(expected):
                    break
            else:
                raise AssertionError(f"{expected} not found in {wdocfiles}")

    def test_sign_text(self):
        signature = self.config.sign_text("hôp")
        self.assertTrue(self.config.check_text_sign("hôp", signature))


if __name__ == "__main__":
    import unittest

    unittest.main()
